import polychromatic.base as base
import polychromatic.common as common
import polychromatic.controller as controller
import polychromatic.locales as locales
import polychromatic.preferences as preferences
import polychromatic.procpid as procpid

import os
import unittest


class TestInternals(unittest.TestCase):
    """
    Test the internals of Polychromatic.
    """
    @classmethod
    def setUpClass(self):
        self.base = base.PolychromaticBase()
        self.base.init_base("", [])

    @classmethod
    def tearDownClass(self):
        pass

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_locales_can_be_set(self):
        i18n = locales.Locales("de_DE")
        _ = i18n.init()
        self.assertEqual(i18n.get_current_locale(), "de_DE", "Could not set up a German locale")

    def test_locales_can_translate_strings(self):
        _ = locales.Locales("de_DE").init()
        # EN: Breath | DE: Atem
        self.assertEqual(_("Breath"), "Atem", "Could not translate text in German")

    def test_locales_can_translate_colours(self):
        _ = locales.Locales("de_DE").init()
        if os.path.exists(self.base.paths.colours):
            os.remove(self.base.paths.colours)
        preferences.init(_)
        colours = preferences.load_file(self.base.paths.colours)
        passed = False
        for item in colours:
            # EN: Green | DE: Grün
            if item["name"] == "Grün":
                passed = True
        self.assertTrue(passed, "Could not translate colour strings")

    def test_config_pref_read(self):
        data = preferences.load_file(self.base.paths.preferences)
        self.assertFalse(data["controller"]["system_qt_theme"], "Could not init or read preferences file")

    def test_config_pref_write(self):
        newdata = preferences.load_file(self.base.paths.preferences)
        newdata["controller"]["landing_tab"] = 2
        preferences.save_file(self.base.paths.preferences, newdata)

        data = preferences.load_file(self.base.paths.preferences)
        self.assertEqual(data["controller"]["landing_tab"], 2, "Could not write to preferences file")

    def test_config_pref_force_invalid_data(self):
        newdata = preferences.load_file(self.base.paths.preferences)
        newdata["controller"]["system_qt_theme"] = 123456
        preferences.save_file(self.base.paths.preferences, newdata)

        # load_file._validate() should correct this
        data = preferences.load_file(self.base.paths.preferences)
        self.assertFalse(data["controller"]["system_qt_theme"], "Invalid data was not corrected")

    def test_data_path(self):
        self.assertTrue(self.base.paths.data_dir.endswith("/data"), "Unexpected development data directory path")

    def test_get_form_factor(self):
        ff = common.get_form_factor(self.base._, "keyboard")
        self.assertEqual(list(ff.keys()), ["id", "icon", "label"], "Unexpected get_form_factor() output")

    def test_get_form_factor_all(self):
        for form_factor_id in common.FORM_FACTORS:
            ff = common.get_form_factor(self.base._, form_factor_id)
        self.assertTrue(True, "Got exception processing form factors")

    def test_get_green_shades(self):
        shades = common.get_green_shades(self.base._)
        passed = True
        for shade in shades:
            if shade["hex"][1:3] != "00" or shade["hex"][5:7] != "00":
                passed = False
        self.assertTrue(passed, "Non-green hex values in get_green_shades()")

    def test_tray_icon_kde(self):
        os.environ["XDG_CURRENT_DESKTOP"] = "KDE"
        os.environ["GTK_THEME"] = "Breeze"
        self.assertEqual(common.get_default_tray_icon(), "img/tray/light/breeze.svg", "Could not detect KDE desktop for tray icon")
        del(os.environ["XDG_CURRENT_DESKTOP"])
        del(os.environ["GTK_THEME"])

    def test_tray_icon_ubuntu(self):
        self.assertEqual(common.get_default_tray_icon(), "img/tray/light/polychromatic.svg", "Coul not retrieve default tray icon")

    def test_get_icon(self):
        self.assertIsNotNone(common.get_icon("general", "controller"), "Could not retrieve an icon")

    def test_colour_bitmap(self):
        self.assertIsNotNone(common.generate_colour_bitmap(self.base.dbg, "#00FF00"), "Could not generate a colour bitmap")

    def test_asset_bitmaps(self):
        icons = common.get_icon_styles(self.base.dbg, "general", "controller", "#FF0000", "#00FF00", "#0000FF", "#FFFFFF", "#000000", "#808080")
        self.assertEqual(len(icons), 4, "Could not generate icon bitmap")

    def test_rgb_to_hex(self):
        self.assertEqual(common.rgb_to_hex([0, 255, 0]), "#00FF00", "Could not convert RGB to hex")

    def test_hex_to_rgb(self):
        self.assertEqual(common.hex_to_rgb("#FF00FF"), [255, 0, 255], "Could not convert RGB to hex")

    def test_state_set_effect(self):
        state = procpid.DeviceSoftwareState("POLY000001")
        state.set_effect("Untitled Effect 1", "/path/to/icon", "/path/to/effect.json")
        self.assertEqual(state.get_effect()["name"], "Untitled Effect 1", "Could not set effect state")
        self.assertEqual(state.get_effect()["icon"], "/path/to/icon", "Could not set effect state")
        self.assertEqual(state.get_effect()["path"], "/path/to/effect.json", "Could not set effect state")

    def test_state_clear_effect(self):
        state = procpid.DeviceSoftwareState("POLY000001")
        state.clear_effect()
        self.assertEqual(state.get_effect(), None, "Could not clear effect state")

    def test_state_set_preset(self):
        state = procpid.DeviceSoftwareState("POLY000001")
        state.set_preset("Untitled Effect 1", "/path/to/icon", "/path/to/effect.json")
        self.assertEqual(state.get_preset()["name"], "Untitled Effect 1", "Could not set preset state")
        self.assertEqual(state.get_preset()["icon"], "/path/to/icon", "Could not set preset state")
        self.assertEqual(state.get_preset()["path"], "/path/to/effect.json", "Could not set preset state")

    def test_state_clear_preset(self):
        state = procpid.DeviceSoftwareState("POLY000001")
        state.clear_preset()
        self.assertEqual(state.get_preset(), None, "Could not clear preset state")

    def test_exception_fault_ours(self):
        error = """Traceback (most recent call last):
            File ’/usr/lib/python3.10/site-packages/polychromatic/controller/devices.py’, line 545, in _clicked_effect_button
                option.apply()
            File ’/usr/lib/python3.10/site-packages/polychromatic/backends/openrazer.py’, line 777, in apply
                self._rzone.spectrum(1)
            TypeError: RazerFX.spectrum() takes 1 positional argument but 2 were given"""
        self.assertTrue(common.is_exception_fault_by_app(error))

    def test_exception_fault_theirs(self):
        error = """Traceback (most recent call last):
            File ’/usr/lib/python3.10/site-packages/polychromatic/backends/openrazer.py’, line 757, in apply
                self._rzone.none()
            File ’/usr/lib/python3.10/site-packages/openrazer/client/fx.py’, line 119, in none
                self._lighting_dbus.setNone()
            File ’/usr/lib/python3.10/site-packages/dbus/proxies.py’, line 141, in __call__
                return self._connection.call_blocking(self._named_service,
            File ’/usr/lib/python3.10/site-packages/dbus/connection.py’, line 652, in call_blocking
                reply_message = self.send_message_with_reply_and_block(
            dbus.exceptions.DBusException: org.freedesktop.DBus.Error.ServiceUnknown: The name :1.203 was not provided by any .service files"""
        self.assertFalse(common.is_exception_fault_by_app(error))



# Polychromatic is licensed under the GPLv3.
# Copyright (C) 2020-2024 Luke Horwell <code@horwell.me>
"""
Provides the "middle ground" between each backend and Polychromatic's interfaces.

The code that provides each backend is stored in "backends" directory.

Refer to the online documentation for more details:
https://docs.polychromatic.app/
"""

from . import procpid
from . import common
from .backends._backend import Backend
from .troubleshoot import openrazer as openrazer_troubleshoot


TROUBLESHOOT_MODULES = {
    "openrazer": openrazer_troubleshoot.troubleshoot
}


class Middleman(object):
    """
    The 'middleman' that processes the data between Polychromatic's applications
    by blending all the backends together.
    """
    def __init__(self):
        """
        Stores variables for the sessions.
        """
        # The PolychromaticBase() object, passed to backends later.
        self._base = None

        # List of initialized Backend() objects.
        self.backends = []

        # List of Backend() modules that failed to init().
        self.bad_init = []

        # List of backend string IDs that are not present.
        self.not_installed = []

        # Dictionary of backend IDs referencing troubleshoot() functions, if available.
        #   e.g. "openrazer": TROUBLESHOOT_MODULES.get("openrazer")
        self.troubleshooters = {}

        # Keys containing human readable strings for modules that failed to import.
        #   e.g. "openrazer": "Exception: xyz"
        self.import_errors = {}

        # List of DeviceItem() objects.
        self.device_cache = []

    def init(self):
        """
        Initialise the backend objects. This should be called when the user interface
        is ready. Note that this thread may potentially be blocked if the backend
        hangs while it initialises.
        """
        # Just OpenRazer for now.
        try:
            import polychromatic.backends.openrazer as openrazer_backend
            backend = openrazer_backend.OpenRazerBackend(self._base)
            if backend.init():
                self.backends.append(backend)
            else:
                self.bad_init.append(backend)
        except (ImportError, ModuleNotFoundError):
            self.not_installed.append("openrazer")
        except Exception as e:
            self.import_errors["openrazer"] = common.get_exception_as_string(e)

        try:
            self.troubleshooters["openrazer"] = TROUBLESHOOT_MODULES["openrazer"]
        except NameError:
            # Backend does not have a troubleshooter.
            pass

    def get_backend(self, device):
        """
        Returns the backend object for the specified device.
        """
        # FIXME: Legacy code used this function for "backend_id"
        if type(device) == str:
            backend_id = device
            for backend in self.backends:
                if backend_id == backend.backend_id:
                    return backend

        for module in self.backends:
            if device.backend_id == module.backend_id:
                return module

    def is_backend_running(self, backend_id):
        """
        Returns a boolean to indicate whether a specific backend ID is running
        and was successfully initialized.
        """
        for module in self.backends:
            if module.backend_id == backend_id:
                return True
        return False

    def get_versions(self):
        """
        Return a dictionary of versions for each running backend.
        """
        versions = {}
        for module in self.backends:
            versions[module.backend_id] = module.version
        return versions

    def _reload_device_cache_if_empty(self):
        """
        Reload the cache of DeviceItem()'s if it hasn't been initalized yet.
        """
        if self.device_cache:
            return

        for backend in self.backends:
            device_list = backend.get_devices()

            if not type(device_list) == list:
                continue

            # Assign 'backend' variable into device object
            for device in device_list:
                device.backend = backend

            self.device_cache = self.device_cache + device_list

    def invalidate_cache(self):
        """
        A fault was detected with the device list. For example, a backend's daemon died,
        or devices were inserted/removed.
        """
        self.device_cache = []

    def reload_device_cache(self):
        """
        Clear the device object cache and reload.
        """
        self.device_cache = []
        self._reload_device_cache_if_empty()

    def get_devices(self):
        """
        Returns a list of DeviceItem() objects.
        """
        self._reload_device_cache_if_empty()
        return self.device_cache

    def get_device_by_name(self, name):
        """
        Returns a fresh DeviceItem() by looking up its device name, or None if
        there is no device with that name.
        """
        for backend in self.backends:
            device = backend.get_device_by_name(name)
            if isinstance(device, Backend.DeviceItem):
                return device
        return None

    def get_device_by_serial(self, serial):
        """
        Returns a fresh DeviceItem() object by looking up its serial number, or
        None if there is no device with that serial string.
        """
        for backend in self.backends:
            device = backend.get_device_by_serial(serial)
            if isinstance(device, Backend.DeviceItem):
                return device
        return None

    def get_devices_by_form_factor(self, form_factor_id):
        """
        Returns a list of DeviceItem()'s based on the form factor specified, or empty list.
        """
        self._reload_device_cache_if_empty()
        devices = []
        for device in self.device_cache:
            if device.form_factor["id"] == form_factor_id:
                devices.append(device)
        return devices

    def get_unsupported_devices(self):
        """
        Returns a list of connected devices that cannot be controlled by their backend.
        """
        unknown_devices = []
        for backend in self.backends:
            device_list = backend.get_unsupported_devices()

            # Assign 'backend' variable into device object
            for device in device_list:
                device.backend = backend

            unknown_devices = unknown_devices + device_list
        return unknown_devices

    def troubleshoot(self, backend, i18n, fn_progress_set_max, fn_progress_advance):
        """
        Performs a series of troubleshooting steps to identify possible
        reasons why a particular backend is non-functional.

        Params:
            backend         (str)       ID of backend to check
            i18n            (obj)       _ function for translating strings
            fn_progress_set_max         See _backend.Backend.troubleshoot()
            fn_progress_advance         See _backend.Backend.troubleshoot()

        Returns:
            (list)          Results from the troubleshooter
            None            Troubleshooter not available
            False           Troubleshooter failed
        """
        try:
            return self.troubleshooters[backend](i18n, fn_progress_set_max, fn_progress_advance)
        except KeyError:
            # Troubleshooter not available for this backend
            return None
        # TODO: Catch errors via interfaces

    def restart(self, backend):
        """
        Restarts a specific backend.
        """
        for module in self.backends:
            if module.backend_id == backend:
                return module.restart()

    def get_active_effect(self, zone):
        """
        Return the first active Backend.EffectOption from the specified zone.
        """
        for option in zone.options:
            if isinstance(option, Backend.EffectOption) and option.active:
                return option

    def get_active_parameter(self, option):
        """
        Return the active Backend.Option.Parameter from the specified option.
        """
        for param in option.parameters:
            if param.active:
                return param

    def get_active_colours_required(self, option):
        """
        Return the number of colours required for the specified option.
        When parameters are present, there may be a different number of colours.
        """
        param = self.get_active_parameter(option)
        if param:
            return param.colours_required
        return option.colours_required

    def get_default_parameter(self, option):
        """
        Return the default Backend.Option.Parameter() object for an option.

        There should only be one default, so the first one will be returned.
        If there are no defaults, the first parameter will be returned.
        """
        if not option.parameters:
            return None

        for param in option.parameters:
            if param.default:
                return param

        return option.parameters[0]

    def _apply_option_with_same_params(self, option):
        """
        Re-apply the specified Backend.Option() instance, using the same
        parameters and colours.
        """
        if option.parameters:
            param_data = option.parameters[0].data
            for param in option.parameters:
                if param.active:
                    param_data = param.data
            option.apply(param_data)

        elif isinstance(option, Backend.ToggleOption):
            option.apply(option.active)

        elif isinstance(option, Backend.SliderOption):
            option.apply(option.value)

        elif isinstance(option, (Backend.EffectOption, Backend.MultipleChoiceOption)):
            option.apply()

    def replay_active_effect(self, device):
        """
        Re-applies the 'active' effect for all zones on the device.

        For example, this may be used to restore previously played effect prior to
        opening the effect editor which was physically previewing on the hardware.
        """
        # TODO: Catch error?
        device.refresh()

        # Was the device playing a software effect?
        state = procpid.DeviceSoftwareState(device.serial)
        effect = state.get_effect()
        if effect:
            procmgr = procpid.ProcessManager("helper")
            procmgr.start_component(["--run-fx", effect["path"], "--device-serial", device.serial])
            return

        # Was the device running a hardware effect?
        for zone in device.zones:
            option = self.get_active_effect(zone)
            if option:
                if option.active:
                    self._apply_option_with_same_params(option)

    def set_colour_for_option(self, option, hex_value, colour_pos=0):
        """
        Set a new colour for the specified option.

        Params:
            option      (obj)   Backend.Option() inherited object
            hex_value   (str)   New #RRGGBB string
            colour_pos  (int)   (Optional) Position to append. 0 = Primary, 1 = Secondary, etc
        """
        option.colours[colour_pos] = hex_value
        self._apply_option_with_same_params(option)

    def set_colour_for_active_effect_zone(self, zone, hex_value, colour_pos=0):
        """
        Set a new colour for the effect that's active in the specified zone.

        Params:
            zone        (obj)   Backend.DeviceItem.Zone() object
            hex_value   (str)   New #RRGGBB string
            colour_pos  (int)   (Optional) Position to append. 0 = Primary, 1 = Secondary, etc
        """
        option = self.get_active_effect(zone)
        if option:
            return self.set_colour_for_option(option, hex_value, colour_pos)

    def set_colour_for_active_effect_device(self, device, hex_value, colour_pos=0):
        """
        Set a new colour for all the device's active effects.

        Params:
            device      (obj)   Backend.DeviceItem() object
            hex_value   (str)   New #RRGGBB string
            colour_pos  (int)   (Optional) Position to append. 0 = Primary, 1 = Secondary, etc
        """
        for zone in device.zones:
            option = self.get_active_effect(zone)
            if option:
                self.set_colour_for_option(option, hex_value, colour_pos)

    def stop_software_effect(self, serial):
        """
        Prior to applying a hardware effect, make sure any software effects
        have stopped.
        """
        process = procpid.ProcessManager(serial)
        state = procpid.DeviceSoftwareState(serial)

        if state.get_effect() or process.is_another_instance_is_running():
            process.stop()
            state.clear_effect()

        if state.get_preset():
            state.clear_preset()



# Polychromatic is licensed under the GPLv3.
# Copyright (C) 2020-2024 Luke Horwell <code@horwell.me>
"""
Provides a common class for handling save data across features of the
application that utilise individual JSON files, such as effects and presets.
"""

import glob
import json
import os
import shutil

from .base import PolychromaticBase as PolychromaticBase
from . import common
from .preferences import VERSION as VERSION

# Error Codes
ERROR_NEWER_FORMAT = -100
ERROR_BAD_DATA = -101
ERROR_MISSING_FILE = -102


class FlatFileManagement(PolychromaticBase):
    """
    Provides common functions for parsing a flat file structure of JSON files.
    This is used for storage for the application's feature offerings, such as
    effects and presets.
    """
    def __init__(self):
        self.feature = "unknown"

        # Paths to where save data can be found for this feature.
        # "factory" refers to files the application ships by default.
        # "local" refers to files that the user created.
        self.factory_path = ""
        self.local_path = ""

    def _get_safe_filename(self, filename):
        """
        Returns a compatible, safe path for the specified filename.
        """
        return "".join([c for c in filename if c.isalpha() or c.isdigit() or c==' ']).replace(" ", "-").strip()

    def _get_file_list(self):
        """
        Returns a list of JSON files.

        Each item consists of a dictionary:
        {
            "name": (str, localized),
            "icon": (str, absolute path),
            "path": (str, path to file),
            "type": (int, feature specific definition - e.g. effect type),
            "editable: (bool, can file be edited)
        }
        """
        file_list = []
        file_list += glob.glob(self.factory_path + "/*.json")
        file_list += glob.glob(self.local_path + "/*.json")
        return file_list

    def _load_file(self, file_path):
        """
        Load the JSON file into memory.

        Returns a dictionary containing the data, or None on failure.
        """
        data = {}

        if not os.path.exists(file_path):
            self.dbg.stdout("{0} no longer exists: {1}".format(self.feature.capitalize(), file_path), self.dbg.error)
            return None

        try:
            with open(file_path, "r") as f:
                data = json.load(f)
        except json.decoder.JSONDecodeError as e:
            self.dbg.stdout("{0} contains invalid JSON: {1}\n{2}".format(self.feature.capitalize(), file_path, str(e)), self.dbg.error)
            return None

        return data

    def _get_i18n_key(self, data, key):
        """
        Returns a localized key if available.
        """
        locale_name = self.i18n.get_current_locale()
        localized_key_full = key + "_" + locale_name
        localized_key_partial = key + "_" + locale_name[:2]

        if localized_key_full in data.keys():
            return data[localized_key_full]

        if localized_key_partial in data.keys():
            return data[localized_key_partial]

        return data[key]

    def _get_icon(self, icon_path):
        """
        Returns the absolute path to the icon for use in the user interface.

        If the icon cannot be loaded, a fallback will be used.
        """
        possible_paths = []

        # It could already be an absolute path
        possible_paths.append(icon_path)

        # Try relative (custom) icons
        possible_paths.append(os.path.join(self.paths.custom_icons, icon_path))

        # Try relative (built-in) icons
        possible_paths.append(os.path.join(self.paths.data_dir, icon_path))

        for path in possible_paths:
            if os.path.exists(path):
                return path

        self.dbg.stdout("Could not locate suitable icon: '{0}'".format(icon_path), self.dbg.warning, 1)
        return common.get_icon("emblems", "software")

    def _validate_key(self, data, key, data_type=None):
        """
        Returns a boolean to indicate whether the key has a value and matches the expected data type.

        Not specifying "data_type" will just check the key exists.
        """
        try:
            data[key]
        except KeyError:
            self.dbg.stdout("{0} file missing required key: {1} ({2})".format(self.feature.capitalize(), key, data_type.__name__), self.dbg.error)
            return False

        if not data_type:
            return True

        if type(data[key]) == data_type:
            return True

        return False

    def _get_parsed_keys(self, data, path):
        """
        Returns a dictionary with parsed values of the specified file. This is
        for use with user interface elements, like titles or lists.

        For example, this would be used to show the localised name and icon of
        an effect, as well as internal data like whether it's editable, the file path
        and effect type.
        """
        parsed = {}
        parsed["name"] = self._get_i18n_key(data, "name")
        parsed["icon"] = self._get_icon(data["icon"])
        parsed["type"] = data["type"]
        parsed["editable"] = os.access(path, os.W_OK)
        parsed["path"] = path
        return parsed

    def get_item_list(self):
        """
        Returns a list of parsed files for use with UI interaction, e.g. effect list
        on sidebar or CLI.
        """
        self.dbg.stdout("Loading list of {0}...".format(self.feature), self.dbg.action, 1)
        file_list = self._get_file_list()
        items = []

        for path in file_list:
            data = self._load_file(path)
            if not data:
                continue
            try:
                items.append(self._get_parsed_keys(data, path))
                self.dbg.stdout("- " + path, self.dbg.action, 1)
            except KeyError as e:
                self.dbg.stdout("Skipping invalid {0} file: {1}".format(self.feature, path), self.dbg.warning)

        return items

    def get_item_list_by_key_filter(self, key, value):
        """
        Return a list of parsed files filtered by the specified key (if exists).

        For example, check the "map_device" key and only return a specific device.
        """
        items = []
        file_list = self._get_file_list()

        for path in file_list:
            data = self._load_file(path)
            try:
                if data[key] == value:
                    items.append(self._get_parsed_keys(data, path))
            except KeyError:
                # Key does not exist or matched
                pass

        return items

    def get_item(self, path):
        """
        Load the item into memory and ensure the data is valid.
        The inheriting class should implement this accordingly.

        Returns:
            {}          Data for the feature as defined by the documentation.
            ERROR_*     One of ERROR_* variables in the root of this module.
        """
        raise NotImplementedError

    def save_item(self, data, orig_path=None):
        """
        Save JSON data to disk.

        Params:
            data        (dict)      Save data for this feature
            orig_path   (str)       Optional. Path referencing where the file was loaded from

        Returns a tuple:
            (success, new_path)     success  = Boolean to indicate success or failure.
                                    new_path = Path to the file's new path (if this has changed)
        """
        # Determine the file name that is consistent with the save data's name.
        filename = self._get_safe_filename(data["name"].replace(" ", "-"))
        target_path = os.path.join(self.local_path, filename + ".json")
        has_script = False

        if orig_path and os.path.exists(orig_path.replace(".json", ".py")):
            has_script = True

        # File is new or path changed, make sure there is no duplicate
        if orig_path != target_path:
            suffix = 2
            while os.path.exists(target_path):
                target_path = os.path.join(self.local_path, filename + "-" + str(suffix) + ".json")
                suffix += 1

            self.dbg.stdout("Name changed. File will be written to: " + target_path, self.dbg.action, 1)

        # Don't save parsed (temporary) data in memory
        if "parsed" in data.keys():
            del(data["parsed"])

        # Save the file!
        try:
            with open(target_path, "w+") as f:
                f.write(json.dumps(data, indent=4))
        except Exception as e:
            self.dbg.stdout("Save Failed: " + target_path, self.dbg.error)
            self.dbg.stdout(common.get_exception_as_string(e), self.dbg.error)
            return (False, target_path)

        # If the file was renamed, only delete if the save succeeded.
        if orig_path and orig_path != target_path:
            os.remove(orig_path)
            self.dbg.stdout("Old filename deleted: " + orig_path, self.dbg.success, 1)

            if has_script:
                orig_script_path = orig_path.replace(".json", ".py")
                dest_script_path = target_path.replace(".json", ".py")
                shutil.move(orig_script_path, dest_script_path)
                self.dbg.stdout("Effect script relocated: " + dest_script_path, self.dbg.success, 1)

        self.dbg.stdout("Save OK: " + target_path, self.dbg.success, 1)
        return (True, target_path)

    def init_data(self, item_name, item_type=None):
        """
        Initialise new blank data for this feature.
        The inheriting class should implement this accordingly.

        Returns (dict) containing new data.
        """
        raise NotImplementedError

    def upgrade_item(self, data):
        """
        Upgrades the data if it was saved in a older application version.
        The inheriting class should implement this accordingly.

        Returns the new data.
        """
        raise NotImplementedError

    def delete_item(self, path):
        """
        Processes the deletion of an item. This will verify and drop references
        from other files, such as a preset.

        Returns a boolean to indicate success.
        """
        try:
            os.remove(path)
        except Exception as e:
            self.dbg.stdout("Delete Failed: {0}\n{1}".format(path, common.get_exception_as_string(e)), self.dbg.error)

        if os.path.exists(path):
            return False

        self.dbg.stdout("Deleted: " + path, self.dbg.success, 1)
        return True

    def clone_item(self, path):
        """
        Processes the duplication of an item. This will append '(Copy)' for
        the user to modify later.

        Returns:
            (str)           Success: Path to the new file
            None            Failed to clone file
        """
        try:
            data = self._load_file(path)

            if not data:
                return None

            # Append "(Copy)" to name and "-copy" to filename
            data["name"] = self._("[] (Copy)").replace("[]", data["name"])
            new_path = path

            success, new_path = self.save_item(data)
            if success:
                self.dbg.stdout("Clone OK: " + new_path, self.dbg.success, 1)
                return new_path

        except Exception as e:
            self.dbg.stdout("Clone Failed: {0}\n{1}".format(path, common.get_exception_as_string(e)), self.dbg.error)

        return None



import polychromatic.fx as fx

import unittest


class TestFX(unittest.TestCase):
    """
    Test the FX 'helper' API calls for effects to use. Excludes device-specific features.
    """
    @classmethod
    def setUpClass(self):
        self.fx = fx.FX()

    @classmethod
    def tearDownClass(self):
        pass

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_rgb_to_hex(self):
        self.assertEqual(self.fx.rgb_to_hex(64, 128, 255), "#4080FF", "Cannot convert RGB to HEX")

    def test_hex_to_rgb(self):
        self.assertEqual(self.fx.hex_to_rgb("#ff8040"), [255, 128, 64], "Cannot convert HEX to RGB")

    def test_saturate_hex(self):
        # Saturate dull green to a brighter green
        self.assertEqual(self.fx.saturate_hex("#408040", 1).upper(), "#00C000", "Cannot saturate HEX value")

    def test_saturate_rgb(self):
        # Same as test above
        self.assertEqual(self.fx.saturate_rgb([64, 128, 64], 1), [0, 192, 0], "Cannot saturate RGB value")

    def test_hue_hex(self):
        # Change hue of red to the middle of the HSL colour graph (aqua)
        self.assertEqual(self.fx.hue_hex("#FF0000", 0.5).upper(), "#00FFFF", "Cannot hue HEX value")

    def test_hue_rgb(self):
        # Same as test above
        self.assertEqual(self.fx.hue_rgb([255, 0, 0], 0.5), [0, 255, 255], "Cannot hue RGB value")

    def test_lightness_hex(self):
        # Reduce lightness of black to grey
        self.assertEqual(self.fx.lightness_hex("#FFFFFF", -0.5).upper(), "#7F7F7F", "Cannot set lightness of HEX value")

    def test_lightness_rgb(self):
        # Similar test to above
        self.assertEqual(self.fx.lightness_rgb([0, 0, 0], 0.5), [127, 127, 127], "Cannot set lightness of RGB value")

    def test_gradient_2_colours(self):
        # Gradient from black to white, across 10 steps. Midpoint should be grey.
        gradient = self.fx.gradient(["#000000", "#FFFFFF"], 3)
        self.assertEqual(gradient[1].upper(), "#7F7F7F", "Cannot verify gradient is accurate")

    def test_gradient_3_colours(self):
        # Gradient from black to white to black, across 10 steps. Midpoint should be white.
        gradient = self.fx.gradient(["#000000", "#FFFFFF", "#000000"], 6)
        self.assertEqual(gradient[2].upper(), "#FFFFFF", "Cannot verify gradient is accurate")



# Polychromatic is licensed under the GPLv3.
# Copyright (C) 2020-2024 Luke Horwell <code@horwell.me>
"""
Troubleshooter for OpenRazer >2.0 and 3.x series.

Users occasionally may end up with installation problems due to the nature of
having a driver module and requirement of being in the 'plugdev' group.

Troubleshooting this aims to inform the user and prompt some guidelines to
get the system up and running again.

The future of OpenRazer aims to move to userspace, which will eliminate
a lot of the common driver issues.
"""

import glob
import grp
import requests
import os
import subprocess
import shutil
import sys

from .. import common
from ..backends import _backend

try:
    from openrazer import client as rclient
    PYTHON_LIB_PRESENT = True
    PYTHON_LIB_WORKING = True
except ModuleNotFoundError:
    PYTHON_LIB_PRESENT = False
    PYTHON_LIB_WORKING = False
except (ImportError, Exception):
    PYTHON_LIB_PRESENT = True
    PYTHON_LIB_WORKING = False

OPENRAZER_MODULES = ["razerkbd", "razermouse", "razeraccessory", "razerkraken"]

def __get_razer_usb_pids():
    razer_usb_pids = []
    vendor_files = glob.glob("/sys/bus/usb/devices/*/idVendor")
    for vendor in vendor_files:
        with open(vendor, "r") as f:
            vid = str(f.read()).strip().upper()
            if vid == "1532":
                with open(os.path.dirname(vendor) + "/idProduct") as f:
                    pid = str(f.read()).strip().upper()
                    razer_usb_pids.append(pid)
    return razer_usb_pids


def _is_daemon_installed(_):
    return {
        "test_name": _("Daemon is installed"),
        "suggestions": [
            _("Install the 'openrazer-meta' package for your distribution.")
        ],
        "passed": True if shutil.which("openrazer-daemon") != None else False
    }


def _is_daemon_running(_):
    try:
        daemon_pid_file = os.path.join(os.environ["XDG_RUNTIME_DIR"], "openrazer-daemon.pid")
    except KeyError:
        daemon_pid_file = os.path.join("/run/user/", str(os.getuid()), "openrazer-daemon.pid")

    daemon_running = False
    if os.path.exists(daemon_pid_file):
        with open(daemon_pid_file) as f:
            daemon_pid = int(f.readline())
        if os.path.exists("/proc/" + str(daemon_pid)):
            daemon_running = True

    return {
        "test_name": _("Daemon is running"),
        "suggestions": [
            _("Start the daemon from the terminal. Run this command and look for errors:"),
            "$ openrazer-daemon -Fv",
        ],
        "passed": daemon_running
    }


def _is_pylib_installed(_):
    py_version = sys.version.split(" ")[0]
    custom_install = not sys.executable.startswith("/usr/bin")

    if PYTHON_LIB_PRESENT and not PYTHON_LIB_WORKING:
        suggestions = [_("Python library is installed, but there is an import error.")]
        if custom_install:
            suggestions.append(_("Python installation looks custom. Check your dependencies are installed correctly."))
        else:
            suggestions.append(_("Check your system is fully up-to-date. Dependencies may be causing problems."))
            suggestions.append(_("For full details on the error, type into a terminal:"))
            suggestions.append("$ python -c \"import openrazer.client\"")

        return {
            "test_name": _("Python library is installed"),
            "suggestions": suggestions,
            "passed": False
        }

    suggestions = [_("Install 'python-openrazer' or 'python3-openrazer' for your distribution.")]

    # Is a custom Python installation interfering?
    if custom_install:
        suggestions.append(_("Your Python installation looks custom. Check your PATH and PYTHONPATH."))

    # Was it previously installed for an older version of Python?
    python_version_major = sys.version.split(".")[0]
    python_version_minor = int(sys.version.split(".")[1])
    for minor_version in range(python_version_minor - 1, 0, -1):
        path = f"/usr/lib/python{python_version_major}.{minor_version}/site-packages/openrazer"
        if os.path.exists(os.path.join(path, "__init__.py")):
            suggestions.append(_("Found library for an older Python version at /usr/lib. Try re-installing.").replace("/usr/lib", path))

    suggestions.append(_("Currently running Python 3.10.0 from /usr/bin/python3").replace("3.10.0", py_version).replace("/usr/bin/python3", sys.executable))

    for path in sys.path:
        suggestions.append("  -- " + _("Tried path:") + " " + path)

    return {
        "test_name": _("Python library is installed"),
        "suggestions": suggestions,
        "passed": PYTHON_LIB_PRESENT
    }


def _is_driver_src_installed(_):
    if not PYTHON_LIB_WORKING:
        return {
            "test_name": _("Kernel module sources are installed for DKMS"),
            "suggestions": [
                _("Unable to check because the Python library is not available.")
            ],
            "passed": None
        }

    if not os.path.exists("/var/lib/dkms"):
        return {
            "test_name": _("Kernel module sources are installed for DKMS"),
            "suggestions": [
                _("OpenRazer is a kernel module, typically built using DKMS."),
                _("Install the 'openrazer-driver-dkms' package for your distribution."),
                _("Some distributions (like Gentoo) may use an alternate system to DKMS."),
            ],
            "passed": None
        }

    driver_version = rclient.__version__
    expected_src_path = "/var/lib/dkms/openrazer-driver/{0}".format(driver_version)

    return {
        "test_name": _("Kernel module sources are installed for DKMS"),
        "suggestions": [
            _("OpenRazer is a kernel module, typically built using DKMS."),
            _("Install the 'openrazer-driver-dkms' package for your distribution."),
        ],
        "passed": True if os.path.exists(expected_src_path) else False
    }


def _is_driver_built(_):
    uname = os.uname()
    kernel_version = uname.release
    driver_version = "<version>"

    if PYTHON_LIB_WORKING:
        driver_version = rclient.__version__

    modules = glob.glob("/lib/modules/{0}/**/razer*.ko*".format(kernel_version), recursive=True)
    modules += glob.glob("/usr/lib/{0}/**/razer*.ko*".format(kernel_version), recursive=True)

    return {
        "test_name": _("Kernel module has been built for this kernel"),
        "suggestions": [
            _("Ensure you have the correct Linux kernel headers package installed for your distribution."),
            _("Your distro's package system might not have rebuilt the module for this kernel. Try running:"),
            "$ sudo dkms install -m openrazer-driver/x.x.x".replace("x.x.x", driver_version),
        ],
        "passed": len(modules) >= len(OPENRAZER_MODULES)
    }


def _can_driver_be_probed(_):
    try:
        results = []
        for module in OPENRAZER_MODULES:
            modprobe = subprocess.Popen(["modprobe", "-n", module], stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
            output = modprobe.communicate()[0].decode("utf-8")
            results.append(True if modprobe.returncode == 0 else False)

        return {
            "test_name": _("Kernel module can be probed"),
            "suggestions": [
                "{0} {1}".format(_("OpenRazer uses these modules:"), ", ".join(OPENRAZER_MODULES)),
                _("For full error details, run this and substitute 'razerkbd' as necessary:"),
                "$ sudo modprobe razerkbd",
                _("No output from this command indicates success."),
            ],
            "passed": True if True in results else False
        }
    except FileNotFoundError:
        return {
            "test_name": _("Kernel module can be probed"),
            "suggestions": [
                _("Could not check as '[]' is not installed.").replace("[]", "modprobe"),
                "{0} {1}".format(_("OpenRazer uses these modules:"), ", ".join(OPENRAZER_MODULES)),
            ],
            "passed": None
        }


def _is_driver_loaded(_):
    try:
        lsmod = subprocess.Popen(["lsmod"], stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
        output = lsmod.communicate()[0].decode("utf-8")
    except FileNotFoundError:
        return {
            "test_name": _("Kernel module is loaded"),
            "suggestions": [
                "{0} {1}".format(_("OpenRazer uses these modules:"), ", ".join(OPENRAZER_MODULES)),
                _("Could not check as '[]' is not installed.").replace("[]", "lsmod"),
            ],
            "passed": None
        }

    return {
        "test_name": _("Kernel module is loaded"),
        "suggestions": [
            "{0} {1}".format(_("OpenRazer uses these modules:"), ", ".join(OPENRAZER_MODULES)),
            _("For full error details, run this and substitute 'razerkbd' as necessary:"),
            "$ sudo modprobe razerkbd",
            _("No output from this command indicates success."),
            _("If you've just installed, it is recommended to restart the computer."),
        ],
        "passed": True if output.find("razer") != -1 else False
    }


def _is_secure_boot_enabled(_):
    sb_sysfile = glob.glob("/sys/firmware/efi/efivars/SecureBoot*")
    sb_reason = _("Secure Boot prevents the driver from loading, as OpenRazer's kernel modules built by DKMS are usually unsigned.")
    passed = False

    # If the modules are loaded, secure boot isn't a problem
    try:
        # Copy of _is_driver_loaded()
        lsmod = subprocess.Popen(["lsmod"], stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
        output = lsmod.communicate()[0].decode("utf-8")
        if output.find("razer") >= 0:
            passed = True
    except FileNotFoundError:
        pass

    if len(sb_sysfile) > 0:
        # The last digit of this sysfs file indicates whether secure boot is enabled
        secureboot = subprocess.Popen(["od", "--address-radix=n", "--format=u1", sb_sysfile[0]], stdout=subprocess.PIPE)
        status = secureboot.communicate()[0].decode("utf-8").split(" ")[-1].strip()

        if int(status) == 0:
            passed = True

        return {
            "test_name": _("Check Secure Boot (EFI) status"),
            "suggestions": [
                _("Secure boot is enabled. Turn it off in the system's EFI settings or sign the modules yourself."),
                sb_reason,
            ],
            "passed": passed
        }

    # Possibly "invalid argument". Can't be sure if it's on or off.
    return {
        "test_name": _("Check Secure Boot (EFI) status"),
        "suggestions": [
            _("Unable to automatically check. If it's enabled, turn it off in the system's EFI settings or sign the modules yourself."),
            sb_reason,
        ],
        "passed": None
    }


def _is_user_in_plugdev_group(_):
    return {
        "test_name": _("User account has been added to the 'plugdev' group"),
        "suggestions": [
            _("Run this command, log out, then log back in to the computer:"),
            "$ sudo gpasswd -a $USER plugdev",
            _("If you've just installed, it is recommended to restart the computer."),
            _("This is required so that your user account (and daemon) has permission to access the driver files in /sys/bus/hid/drivers"),
        ],
        "passed": _backend.BackendHelpers().is_user_in_group("plugdev")
    }


def _is_sysfs_plugdev_permissions_ok(_):
    if not os.path.exists("/sys/bus/hid/drivers/"):
        return {
            "test_name": _("Device can be accessed using plugdev permissions"),
            "suggestions": [
                _("Unable to check as the drivers sysfs path for this Linux kernel is non-standard."),
            ],
            "passed": None
        }

    razer_modules = glob.glob("/sys/bus/hid/drivers/razer*")
    if not razer_modules:
        return {
            "test_name": _("Device can be accessed using plugdev permissions"),
            "suggestions": [
                _("Unable to check because OpenRazer's modules are not loaded."),
            ],
            "passed": None
        }

    # Gather file list
    sysfs_files = []
    for driver in razer_modules:
        for prefix in ["device_", "matrix_"]:
            sysfs_files += glob.glob("{0}/*/{1}*".format(driver, prefix))

    # Check sample of sysfs files have the correct group permission
    results = []
    for sysfile in sysfs_files:
        gid = os.stat(sysfile).st_gid
        group_name = grp.getgrgid(gid)[0]
        results.append(True if group_name == "plugdev" else False)

    return {
        "test_name": _("Device can be accessed using plugdev permissions"),
        "suggestions": [
            _("Permissions for OpenRazer's device files (/sys/bus/hid/drivers) are not set correctly. They should be owned as root:plugdev (owner/group)"),
            _("Replugging the hardware and then restarting the daemon usually fixes the problem."),
            _("Alternately, improperly configured udev rules (such as a missing PID) may cause this to happen."),
        ],
        "passed": False if False in results else True
    }


def _is_razer_device_connected(_):
    razer_usb_pids = __get_razer_usb_pids()

    return {
        "test_name": _("Razer hardware is present for kernel"),
        "suggestions": [
            _("The Linux kernel does not see hardware with a VID of 1532."),
            _("Check the device and USB port is working. Try a different port."),
        ],
        "passed": len(razer_usb_pids) > 0
    }


def _check_device_support_list(_):
    remote_get_url = "https://openrazer.github.io/api/devices.json"
    remote_device_list = None
    missing_pids = []

    try:
        request = requests.get(remote_get_url)
        remote_device_list = request.json()
    except Exception as e:
        # Gracefully ignore connection errors
        print("Could not retrieve OpenRazer data: {0}\n{1}".format(remote_get_url, str(e)))

    def _remote_failed():
        return {
            "test_name": _("Check device compatibility"),
            "suggestions": [
                _("Unable to retrieve this data from OpenRazer's website."),
                _("Check the OpenRazer website to confirm your device is listed as supported."),
                _("If you're checking OpenRazer's GitHub repository, check the 'stable' branch."),
            ],
            "passed": None
        }

    if not remote_device_list:
        return _remote_failed()

    # Get list of USB VIDs and PIDs plugged into the system.
    razer_usb_pids = __get_razer_usb_pids()

    # Check Razer PIDs against remote device list
    try:
        remote_pids = []
        for device in remote_device_list:
            remote_pids.append(device["pid"])
            for alias in device.get("alias_ids", []):
                remote_pids.append(alias[5:])
    except KeyError as e:
        print("Remote device list contains invalid data: " + remote_get_url)
        return _remote_failed()

    for pid in razer_usb_pids:
        if not pid in remote_pids:
            missing_pids.append("1532:" + pid)

    return {
        "test_name": _("Check device compatibility"),
        "suggestions": [
            _("The following PIDs are not supported in the latest stable release of OpenRazer:"),
            ", ".join(missing_pids),
            _("Check OpenRazer's issues and pull requests to see the status for your device. Create an issue if necessary."),
        ],
        "passed": len(missing_pids) == 0
    }


def _is_openrazer_up_to_date(_):
    if PYTHON_LIB_WORKING:
        local_version = rclient.__version__
        remote_version = None
        remote_get_url = "https://openrazer.github.io/api/latest_version.txt"
        try:
            request = requests.get(remote_get_url)

            # Response should be 3 digits, e.g. 3.0.1
            if request.status_code == 200 and len(request.text.strip().split(".")) == 3:
                remote_version = request.text.strip()
        except Exception as e:
            # Gracefully ignore connection errors
            print("Could not retrieve OpenRazer data: {0}\n{1}".format(remote_get_url, str(e)))
            request = None

        def _is_version_newer_then(verA, verB):
            verA = verA.split(".")
            verB = verB.split(".")
            is_new = False

            if int(verA[0]) > int(verB[0]):
                is_new = True

            if float(verA[1] + '.' + verA[2]) > float(verB[1] + '.' + verB[2]):
                is_new = True

            return is_new

        if remote_version:
            return {
                "test_name": _("OpenRazer is the latest version"),
                "suggestions": [
                    _("There is a new version of OpenRazer available."),
                    _("New versions add support for more devices and address device-specific issues."),
                    _("Your version: 0.0.0").replace("0.0.0", local_version),
                    _("Latest version: 0.0.0").replace("0.0.0", remote_version),
                ],
                "passed": _is_version_newer_then(remote_version, local_version) is not True
            }

        return {
            "test_name": _("OpenRazer is the latest version"),
            "suggestions": [
                _("Unable to retrieve this data from OpenRazer's website."),
                _("Check the OpenRazer website to confirm your device is listed as supported."),
                _("If you're checking OpenRazer's GitHub repository, check the 'stable' branch."),
            ],
            "passed": None
        }

    return {
        "test_name": _("OpenRazer is the latest version"),
        "suggestions": [
            _("Unable to check because the Python library is not available.")
        ],
        "passed": None
    }


def troubleshoot(_, fn_progress_set_max, fn_progress_advance):
    """
    See: _backend.Backend.troubleshoot()
    """
    results = []
    fn_progress_set_max(10)

    # Abort if running on a non-Linux system. Troubleshooting was written for Linux.
    if os.uname().sysname != "Linux":
        return None

    try:
        # Perform the checks in a logical order
        for test in [_is_daemon_installed,
                     _is_pylib_installed,
                     _is_driver_src_installed,
                     _is_driver_built,
                     _can_driver_be_probed,
                     _is_driver_loaded,
                     _is_daemon_running]:
            results.append(test(_))
            fn_progress_advance()

        # EFI systems only
        if os.path.exists("/sys/firmware/efi"):
            results.append(_is_secure_boot_enabled(_))
        fn_progress_advance()

        for test in [_is_razer_device_connected,
                     _is_user_in_plugdev_group,
                     _is_sysfs_plugdev_permissions_ok,
                     _check_device_support_list,
                     _is_openrazer_up_to_date]:
            results.append(test(_))
            fn_progress_advance()

    except Exception as e:
        exception = common.get_exception_as_string(e)
        print("Failed to run OpenRazer troubleshooter!\n" + exception)
        return exception

    return results



# Polychromatic is licensed under the GPLv3.
# Copyright (C) 2021-2024 Luke Horwell <code@horwell.me>
"""
Variables and functions for essential operation of the application.
All interfaces and most classes will inherit these.
"""
from .locales import Locales
from .middleman import Middleman

from .paths import Paths as Paths
from . import common as common_class
from . import preferences as preferences_class

import os


class PolychromaticBase(object):
    """
    Essential variables/functions for minimum functionality of the application.
    """
    # Should the process restart later
    exec_path = ""
    exec_args = []

    # Localization
    i18n = Locales()
    _ = i18n.init()

    # Storage
    paths = Paths()

    # TODO: Refactor later
    common = common_class
    common.paths = paths
    dbg = common.Debugging()
    pref = preferences_class
    pref.path = paths
    pref.init(_)
    preferences = pref.load_file(paths.preferences)

    # Devices
    middleman = Middleman()

    # Functions
    @classmethod
    def init_base(self, path, args):
        """
        Finish populating variables. Run this after initializing the class.
        """
        PolychromaticBase.exec_path = path
        PolychromaticBase.exec_args = args
        PolychromaticBase.middleman._base = self

    @classmethod
    def reinit_locales(self, locale):
        PolychromaticBase.i18n = Locales(locale)
        PolychromaticBase._ = self.i18n.init()



# Polychromatic is licensed under the GPLv3.
# Copyright (C) 2020-2024 Luke Horwell <code@horwell.me>
"""
This module contains the events when clicking on interactive menu bar components.
"""

import os
import time
import subprocess
import webbrowser

from ..base import PolychromaticBase
from .. import common
from .. import preferences
from .. import procpid
from . import shared
from . import procviewer
from . import troubleshooter

from PyQt5.QtCore import Qt, QSize
from PyQt5.QtGui import QBrush, QPixmap, QFont, QIcon
from PyQt5.QtWidgets import QWidget, QMenuBar, QMenu, QAction, QLabel, QDialog, \
                            QPushButton, QTreeWidget, QTreeWidgetItem, \
                            QTextEdit, QButtonGroup, QProgressBar, QMessageBox, \
                            QApplication, QListWidget, QListWidgetItem


class MenuBar(PolychromaticBase):
    """
    Allows the user to quickly change the existing state of the device right now.
    """
    def __init__(self, appdata):
        self.appdata = appdata
        self._ = appdata._
        self.mainwindow = self.appdata.main_window
        self.menubar = self.mainwindow.findChild(QMenuBar)
        self.widgets = shared.PolychromaticWidgets(appdata)

        # Classes per backend
        self.human_name = None
        self.openrazer = MenuBarOpenRazer(appdata, self.widgets)

        # Bind global menu bar items to their events
        # -- File
        self._bind_item("actionNewEffect", self.new_effect)
        self._bind_item("actionImportEffect", self.import_effect)
        self._bind_item("actionNewPreset", self.new_preset)
        self._bind_item("actionNewPresetNow", self.new_preset_now)

        # -- View
        self._bind_item("actionHideMenuBar", self.hide_menu_bar)
        self._bind_item("actionReinstateMenuBar", self.reinstate_menu_bar)
        self._bind_item("actionForceRefresh", self.force_refresh)
        self._bind_item("actionPreferences", self.open_preferences)

        # -- Tools
        self._bind_item("actionRestartTrayApplet", self.restart_tray_applet)
        self._bind_item("actionProcessViewer", self.open_process_viewer)

        # -- Tools > OpenRazer
        self._bind_item("actionOpenRazerWebsite", self.openrazer.website)
        self._bind_item("actionOpenRazerReportBug", self.openrazer.report_bug)
        self._bind_item("actionOpenRazerReleaseNotes", self.openrazer.release_notes)
        self._bind_item("actionOpenRazerConfigure", self.openrazer.configure)
        self._bind_item("actionOpenRazerOpenLog", self.openrazer.open_log)
        self._bind_item("actionOpenRazerRestartDaemon", self.openrazer.restart_daemon)
        self._bind_item("actionOpenRazerAbout", self.openrazer.about)

        # -- Tools > Troubleshooter
        self._bind_item("actionTroubleshootOpenRazer", self.openrazer.troubleshoot)

        # -- Help
        self._bind_item("actionOnlineHelp", self.online_help)
        self._bind_item("actionWebsite", self.polychromatic_website)
        self._bind_item("actionReleaseNotes", self.polychromatic_release_notes)
        self._bind_item("actionReportBug", self.polychromatic_report_bug)
        self._bind_item("actionDonate", self.polychromatic_donate)
        self._bind_item("actionAbout", self.about_polychromatic)

        self._load_icons()

    def _load_icons(self):
        """
        Load icons for the menu bar if Polychromatic's theme is used.
        """
        if self.appdata.system_qt_theme:
            return

        def _set_icon(object_name, icon_dir, icon_name, object_type=QAction):
            self.mainwindow.findChild(object_type, object_name).setIcon(self.widgets.get_icon_qt(icon_dir, icon_name))

        # -- File
        _set_icon("actionNewEffect", "general", "effects")
        _set_icon("actionImportEffect", "general", "import")
        _set_icon("actionNewPreset", "general", "presets")
        _set_icon("actionNewPresetNow", "devices", "all")
        _set_icon("actionQuitApp", "general", "exit")

        # -- Edit
        _set_icon("actionDuplicate", "general", "clone")
        _set_icon("actionDelete", "general", "delete")
        _set_icon("actionRefreshTab", "general", "refresh")
        _set_icon("actionForceRefresh", "general", "refresh")
        _set_icon("actionPreferences", "general", "preferences")

        # -- View
        _set_icon("actionDevices", "general", "devices")
        _set_icon("actionEffects", "general", "effects")
        _set_icon("actionPresets", "general", "presets")
        _set_icon("actionTriggers", "general", "triggers")
        _set_icon("actionReinstateMenuBar", "general", "pin")
        _set_icon("actionHideMenuBar", "general", "pin")

        # -- Tools
        _set_icon("actionTroubleshootOpenRazer", "emblems", "utility")
        _set_icon("actionRestartTrayApplet", "general", "tray-applet")
        _set_icon("actionProcessViewer", "emblems", "software")

        # -- Tools/OpenRazer
        _set_icon("actionOpenRazerWebsite", "general", "external")
        _set_icon("actionOpenRazerConfigure", "general", "preferences")
        _set_icon("actionOpenRazerOpenLog", "general", "folder")
        _set_icon("actionOpenRazerRestartDaemon", "general", "refresh")

        # -- Help
        _set_icon("actionOnlineHelp", "general", "external")

    def _bind_item(self, object_name="", function=object):
        """
        Set up a menu bar item to run this function when selected.
        """
        action = self.mainwindow.findChild(QAction, object_name)
        action.triggered.connect(function)

    def new_effect(self):
        self.appdata.tab_effects.new_file()

    def import_effect(self):
        self.appdata.tab_effects.import_effect()
        pass

    def new_preset(self):
        print("stub:menubar.new_preset")
        pass

    def new_preset_now(self):
        print("stub:menubar.new_preset_now")
        pass

    def hide_menu_bar(self):
        """
        Hide the menu bar for this session until further notice.
        """
        self.menubar.hide()
        self.mainwindow.findChild(QAction, "actionReinstateMenuBar").setVisible(True)
        self.preferences["controller"]["show_menu_bar"] = False
        preferences.save_file(self.paths.preferences, self.preferences)
        self.widgets.open_dialog(self.widgets.dialog_generic,
                                 self._("Hide Menu Bar"),
                                 self._("The menu bar is now hidden. To temporarily reveal, press Alt."))

    def reinstate_menu_bar(self):
        """
        Keep the menu bar between sessions
        """
        self.mainwindow.findChild(QAction, "actionReinstateMenuBar").setVisible(False)
        self.preferences["controller"]["show_menu_bar"] = True
        preferences.save_file(self.paths.preferences, self.preferences)

    def _is_editor_running(self):
        """
        Returns a boolean to indicate whether an editor is currently running.
        """
        editors = self.appdata.tab_effects.editors
        for editor in editors:
            if editors[editor].alive:
                return True
        return False

    def force_refresh(self):
        """
        Restarts the application execution so backends can be cleanly re-initialised.
        This is useful when devices are inserted/removed.
        """
        if self._is_editor_running():
            self.widgets.open_dialog(self.widgets.dialog_generic,
                                     self._("Force Refresh"),
                                     self._("Please close all editor windows before executing this action."))
            return

        procmgr = procpid.ProcessManager()
        procmgr.restart_self(self.exec_path, self.exec_args)

    def open_preferences(self):
        self.appdata.ui_preferences.open_window()

    def restart_tray_applet(self):
        process = procpid.ProcessManager("tray-applet")
        process.start_component()

    def open_process_viewer(self):
        procviewer.ProcessViewer(self.appdata)

    def _prompt_on_locale_change(self, title):
        """
        Informs the user if the resource they click is only available in English.
        """
        if not self.i18n.get_current_locale().startswith("en"):
            self.widgets.open_dialog(self.widgets.dialog_generic,
                                     title,
                                     self._("Polychromatic is about to open a web page for this resource. Unfortunately, it is only available in English at this time."))

    def online_help(self):
        self._prompt_on_locale_change(self._("Online Help"))

        # For context, open the most relevant documentation
        context = "/"
        tabs = self.mainwindow.findChild(QWidget, "MainTabWidget")
        if tabs:
            current_tab = tabs.currentIndex()
            indexes = {
                0: "/controller/devices/",
                1: "/controller/effects/",
                2: "/controller/presets/",
                3: "/controller/triggers/"
            }
            context = indexes[current_tab]

        webbrowser.open("https://docs.polychromatic.app" + context)

    def polychromatic_website(self):
        self._prompt_on_locale_change(self._("Website"))
        webbrowser.open("https://polychromatic.app/")

    def polychromatic_release_notes(self):
        self._prompt_on_locale_change(self._("What's New?"))
        webbrowser.open("https://polychromatic.app/permalink/latest/")

    def polychromatic_report_bug(self):
        self._prompt_on_locale_change(self._("Report Bug"))
        webbrowser.open("https://polychromatic.app/permalink/bugs/")

    def polychromatic_donate(self):
        self._prompt_on_locale_change(self._("Donate"))
        webbrowser.open("https://polychromatic.app/permalink/donate/")

    def _show_aboutbox(self, title, logo, homepage_url, versions, license_text, links):
        """
        Shows an about dialog box with the specified parameters.

        Params:
            title           (str)   Application title
            logo            (str)   Absolute path to application's logo
            homepage_url    (str)   HTTPS link to the project's homepage
            versions        (list)  Table of strings: [["name": "ver"]]
            license_text    (str)   Show this license text.
            links           (list)  List of links: [["label": "url"]]
        """
        about = shared.get_ui_widget(self.appdata, "about", QDialog)

        # Logo
        about_icon = about.findChild(QLabel, "AppIcon")
        shared.set_pixmap_for_label(about_icon, logo, 96)

        # App Name
        name = about.findChild(QLabel, "AppName")
        name.setText(title)

        # URL under the application title
        url = about.findChild(QLabel, "AppURL")
        def _link(a):
            webbrowser.open(homepage_url)
        url.mousePressEvent = _link
        url.setText(homepage_url)

        # Set text colour if using Polychromatic Qt theme
        if not self.appdata.system_qt_theme:
            name.setStyleSheet("QLabel { color: lime }")
            url.setStyleSheet("QLabel { color: lime }")

        # List versions & dependencies
        tree = about.findChild(QTreeWidget, "VersionTree")
        column = tree.invisibleRootItem()
        for version in versions:
            item = QTreeWidgetItem()
            item.setText(0, version[0])
            item.setText(1, version[1])
            column.addChild(item)

        # Add application links
        links_tab_widget = about.findChild(QWidget, "TabLinks").layout()
        links_list = QListWidget()
        links_list.setIconSize(QSize(16, 16))

        for link in links:
            label = link[0]
            href = link[1]
            item = QListWidgetItem()
            item.setText(label)
            item.setIcon(self.widgets.get_icon_qt("general", "external"))
            item.setData(Qt.ItemDataRole.UserRole, href)
            links_list.addItem(item)
        links_tab_widget.addWidget(links_list)

        def _link_clicked(item: QListWidgetItem):
            links_list.clearSelection()
            links_list.clearFocus()
            webbrowser.open(item.data(Qt.ItemDataRole.UserRole))

        links_list.itemClicked.connect(_link_clicked)

        # Show license text
        license = about.findChild(QTextEdit, "LicenseText")
        license.setPlainText(license_text)

        # Set up the about window
        close = about.findChild(QPushButton, "Close")
        def _close(a):
            about.accept()
        close.clicked.connect(_close)
        if not self.appdata.system_qt_theme:
            close.setIcon(self.widgets.get_icon_qt("general", "close"))

        window_title = ' '.join(sub[:1].upper() + sub[1:] for sub in title.split(' '))
        about.setWindowTitle(self._("About []").replace("[]", window_title))
        about.setWindowIcon(QIcon(logo))
        about.setWindowFlag(Qt.WindowMinimizeButtonHint, False)
        about.setWindowFlag(Qt.WindowMaximizeButtonHint, False)
        about.exec()

    def about_polychromatic(self):
        logo = common.get_icon("logo", "polychromatic")

        with open(os.path.join(self.appdata.paths.data_dir, "licenses/app.txt"), "r") as f:
            license = f.read()

        links = [
            [self._("Documentation"), "https://docs.polychromatic.app/"],
            [self._("Source Code"), "https://polychromatic.app/permalink/src/"],
            [self._("Release Notes"), "https://polychromatic.app/permalink/latest/"],
            [self._("Donate"), "https://polychromatic.app/permalink/donate/"]
        ]

        versions = self.appdata.versions
        self._show_aboutbox("polychromatic {0}".format(versions[0][1]), logo, "https://polychromatic.app", versions, license, links)


class MenuBarOpenRazer(MenuBar):
    """
    Options for the OpenRazer backend under the Tools menu. When the backend is
    not present, there may be a limited set of options.
    """
    # TODO: Refactor later. Move to separate window.
    def __init__(self, appdata, widgets):
        self.human_name = "OpenRazer"
        self.appdata = appdata
        self.widgets = widgets
        self.module = None
        self.running = None
        self.url_website = "https://openrazer.github.io"
        self.url_src = "https://github.com/openrazer/openrazer/"
        self.url_bugs = "https://github.com/openrazer/openrazer/issues"
        self.url_latest = "https://github.com/openrazer/openrazer/releases/latest"

    def _refresh(self):
        """
        Initialises the variables for this class, if not done so already.
        """
        self.module = self.middleman.get_backend("openrazer")
        self.running = True if self.module else False

    def website(self):
        webbrowser.open(self.url_website)

    def report_bug(self):
        webbrowser.open(self.url_bugs)

    def release_notes(self):
        webbrowser.open(self.url_latest)

    def configure(self):
        self.appdata.ui_preferences.openrazer.open_window()

    def open_log(self):
        cmd = None
        try:
            cmd = "xdg-open file://{0}/openrazer/logs/razer.log".format(os.environ["XDG_DATA_HOME"])
        except KeyError:
            pass

        if not cmd:
            try:
                cmd = "xdg-open file://{0}/.local/share/openrazer/logs/razer.log".format(os.environ["HOME"])
            except KeyError:
                cmd = "xdg-open file:///home/$USER/.local/share/openrazer/logs/razer.log"

        print("Running: " + cmd)
        os.system(cmd)

    def restart_daemon(self):
        """
        Restart the OpenRazer daemon, which also requires the application and
        any background processes (including the tray applet) to restart.
        """
        def _reload_openrazer():
            self.appdata.main_window.hide()
            self.loading = shared.get_ui_widget(self.appdata, "loading", QDialog)

            label = self.loading.findChild(QLabel, "Label")
            label.setText(self._("Restarting the backend/application..."))
            self.loading.show()
            common.run_thread(_reload_openrazer_thread)

        def _reload_openrazer_thread():
            self.appdata.middleman.get_backend("openrazer").restart()
            procmgr = procpid.ProcessManager()
            procmgr.restart_all()
            procmgr.restart_self(self.exec_path, self.exec_args)

        self.widgets.open_dialog(self.widgets.dialog_generic,
                                 self._("Reload application?"),
                                 self._("To apply these changes, a daemon process needs to restart. This will also restart Polychromatic. Any unsaved data will be lost."),
                                 buttons=[QMessageBox.Ok, QMessageBox.Cancel],
                                 default_button=QMessageBox.Ok,
                                 actions={QMessageBox.Ok: _reload_openrazer})

    def _get_dbus_version(self):
        """
        For about dialog, get the D-Bus version (used by the OpenRazer driver)
        """
        try:
            output = subprocess.Popen(["dbus-daemon", "--version"], stdout=subprocess.PIPE).communicate()[0]
            return output.decode("UTF-8").split("\n")[0].split(" ")[-1]
        except Exception:
            self.dbg.stdout("Unable to get D-Bus version! Ignoring.", self.dbg.warning)
            return "[Unknown]"

    def _get_dkms_version(self):
        """
        For about dialog, get the DMKS version (used by the OpenRazer driver)
        """
        try:
            output = subprocess.Popen(["dkms", "--version"], stdout=subprocess.PIPE).communicate()[0]
            # Output: "dkms:2.8" or "dkms-2.8.6"
            return output.decode("UTF-8").strip()[5:]
        except Exception:
            self.dbg.stdout("Unable to get DKMS version! Ignoring.", self.dbg.warning)
            return "[Unknown]"

    def troubleshoot(self):
        self.troubleshooter = troubleshooter.TroubleshooterGUI(self.appdata, "openrazer", "OpenRazer")

    def about(self):
        self._refresh()

        # Cannot show details if backend is not running
        if not self.running:
            self.dbg.stdout("Cannot open OpenRazer about dialog as backend is not loaded!", self.dbg.error)
            return

        logo = common.get_icon("logo", "openrazer")
        links = [
            [self._("Website"), self.url_website],
            [self._("Source Code"), self.url_src],
            [self._("Release Notes"), self.url_latest]
        ]

        with open(os.path.join(self.paths.data_dir, "licenses/openrazer.txt"), "r") as f:
            license = f.read()

        versions = [
            ["Daemon", self.module.devman.daemon_version],
            ["Python Library", self.module.version]
        ]

        # Attempt to add dependencies to the version list
        version_dbus = self._get_dbus_version()
        version_dkms = self._get_dkms_version()

        if version_dbus:
            versions.append(["D-Bus", version_dbus])

        if version_dkms:
            versions.append(["DKMS", version_dkms])

        self._show_aboutbox("OpenRazer {0}".format(self.module.version), logo, "https://openrazer.github.io", versions, license, links)



#!/usr/bin/python3
#
# Validates the integrity of JSON files and (if applicable) the paths for
# the files exist.
#

import glob
import os
import json
import sys

files = glob.glob("data/**/*.json", recursive=True)
passed = True

for path in files:
    try:
        with open(path, "r") as f:
            data = json.load(f)
    except Exception as e:
        print("Invalid JSON: " + path)
        print(str(e))
        passed = False
        continue

    if path.endswith("icons.json"):
        for icon in data:
            if not os.path.exists(os.path.join("data", icon)):
                print("Icon does not exist: " + icon)
                passed = False

if passed:
    sys.exit(0)
else:
    sys.exit(1)



#!/usr/bin/python3
#
# Assumes run_daemon.sh is running.
#
# The tests target the latest version of OpenRazer from the 'master' branch.
#
import os
import random
import unittest

# Polychromatic Modules
from polychromatic import base
from polychromatic.backends._backend import Backend
from polychromatic.backends.openrazer import OpenRazerBackend, OpenRazerPersistence, OpenRazerPersistenceFallback

# External
import openrazer.client


class OpenRazerMiddlemanTest(unittest.TestCase):
    """
    Test OpenRazerBackend against a "fake driver" instance of the OpenRazer daemon,
    using all the supported devices for testing the class is implemented correctly.
    """
    @classmethod
    def setUpClass(self):
        self.base = base.PolychromaticBase()
        self.base.init_base("", [])
        self.openrazer = OpenRazerBackend(self.base)

    @classmethod
    def get_rdevice(self, name):
        """
        Directly get an OpenRazer device for testing.
        """
        for rdevice in self.openrazer.devman.devices:
            if rdevice.name == name:
                return rdevice

    @classmethod
    def get_option(self, device, option_uid, zone_id):
        """
        Returns a specific option from a zone.
        """
        for zone in device.zones:
            if zone.zone_id == zone_id:
                for option in zone.options:
                    if option.uid == option_uid:
                        return option
        raise ValueError(f"Could not find option '{option_uid}' in zone '{zone_id}' for {device.name}")

    @classmethod
    def get_device(self, name):
        """
        Returns a specific device by its name.
        """
        device = self.openrazer.get_device_by_name(name)
        if not device:
            raise ValueError("Could not find device: " + name)
        if not isinstance(device, self.openrazer.DeviceItem):
            raise ValueError(f"Could not get device: {name} due to this error:\n{device}")
        return device

    @classmethod
    def tearDownClass(self):
        pass

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_backend_init(self):
        self.assertTrue(self.openrazer.init(), "OpenRazerBackend could not init")

    def test_client_override(self):
        ripple_override_path = os.path.join(self.openrazer.get_backend_storage_path(), "ripple_refresh_rate")

        # Perform what the GUI would do
        with open(ripple_override_path, "w") as f:
            f.write("1")

        # Does it load correctly?
        self.openrazer.load_client_overrides()
        self.assertEqual(self.openrazer.ripple_refresh_rate, 1, "Failed to override a client setting")

    def test_empty_unrecog_list(self):
        self.assertEqual(self.openrazer.get_unsupported_devices(), [], "Unexpected 'unrecognised device' found")

    def test_get_devices(self):
        self.assertGreater(len(self.openrazer.get_devices()), 140, "Failed to get devices")

    def test_get_device_by_name(self):
        device = self.get_device("Razer BlackWidow Chroma")
        self.assertEqual(device.name, "Razer BlackWidow Chroma", "Failed to get a device by name")

    def test_get_device_by_serial(self):
        # XX0000000203 == razerblackwidowchroma.cfg
        device = self.openrazer.get_device_by_serial("XX0000000203")
        self.assertIsNotNone(device, "Failed to get a device by serial")

    def test_device_dpi_max(self):
        device = self.get_device("Razer Mamba Elite")
        self.assertEqual(device.dpi.max, 16000, "Incorrect Max DPI")

    def test_device_dpi_has_stages(self):
        device = self.get_device("Razer Mamba Elite")
        self.assertEqual(len(device.dpi.default_stages), 5, "Failed to list five DPI stages")

    def test_device_dpi_set_get(self):
        device = self.get_device("Razer DeathAdder V2")
        device.dpi.set(20000, 20000)
        device.dpi.refresh()
        self.assertEqual(device.dpi.x, 20000, "Failed to set or get DPI")

    def test_device_dpi_set_get_deathadder35G(self):
        # In <openrazer>/daemon/openrazer_daemon/hardware/mouse.py:
        # AVAILABLE_DPI = [450, 900, 1800, 3500]
        device = self.get_device("Razer DeathAdder 3.5G")
        option = self.get_option(device, "fixed_dpi", "main")

        if not option:
            raise RuntimeError("Expected a FixedDPIOption object")

        option.apply(900)
        option.refresh()

        # The second parameter should now be selected
        self.assertTrue(option.parameters[1].active, "Failed to set or get for a fixed DPI X device")

    def test_device_matrix(self):
        device = self.get_device("Razer BlackWidow Chroma V2")
        if device.matrix.rows != 6:
            raise ValueError("Unexpected rows: " + str(device.matrix.rows))
        if device.matrix.cols != 22:
            raise ValueError("Unexpected columns: " + str(device.matrix.cols))
        device.matrix.set(0, 0, 255, 0, 0)
        device.matrix.set(1, 0, 0, 255, 0)
        device.matrix.set(2, 0, 0, 0, 255)
        device.matrix.draw()
        device.matrix.clear()
        device.matrix.set(0, 1, 255, 255, 255)

    def test_device_matrix_deathstalker(self):
        # This device normally has 12 columns, but we have a virtual one of 6 due to its quirks.
        device = self.get_device("Razer DeathStalker Chroma")
        self.assertEqual(device.matrix.cols, 6, "Failed to create virtual matrix for Razer DeathStalker Chroma")

    def test_device_matrix_deathstalker_apply(self):
        device = self.get_device("Razer DeathStalker Chroma")
        device.matrix.clear()
        device.matrix.set(0, 0, 255, 0, 0)
        device.matrix.draw()

    def test_device_matrix_name(self):
        device = self.get_device("Razer BlackWidow Ultimate 2016")
        self.assertEqual(device.matrix.name, "Razer BlackWidow Ultimate 2016")

    def test_device_matrix_form_factor_id(self):
        device = self.get_device("Razer BlackWidow Chroma")
        self.assertEqual(device.matrix.form_factor_id, "keyboard")

    def test_zone_logo_laptop(self):
        # The second zone (logo) for some Blade models are the laptop lid
        device = self.get_device("Razer Blade Stealth (Late 2016)")
        self.assertEqual(device.zones[1].label, "Laptop Lid", "Failed to set custom zone label")

    def test_form_factor_stand(self):
        device = self.get_device("Razer Base Station Chroma")
        self.assertEqual(device.form_factor["id"], "stand", "Unexpected form factor ID")

    def test_vid_pid(self):
        device = self.get_device("Razer Mamba Tournament Edition")
        self.assertTrue(device.vid == "1532" and device.pid == "0046", "Failed to get VID:PID values")

    def test_monochromatic_keyboard(self):
        device = self.get_device("Razer BlackWidow Ultimate 2016")
        self.assertTrue(device.monochromatic == True, "Device incorrectly reporting monochromatic status")

    def test_non_monochromatic_headset(self):
        device = self.get_device("Razer Kraken Ultimate")
        self.assertTrue(device.monochromatic == False, "Device incorrectly reporting monochromatic status")

    def test_capability_main(self):
        rdevice = self.get_rdevice("Razer Mamba Tournament Edition")
        zone = OpenRazerBackend.DeviceItem.Zone()
        zone.zone_id = "backlight"
        self.assertTrue(self.openrazer._has_zone_capability(rdevice, zone, "brightness"), "Capability incorrectly reported")

    def test_capability_logo(self):
        rdevice = self.get_rdevice("Razer Abyssus Essential")
        zone = OpenRazerBackend.DeviceItem.Zone()
        zone.zone_id = "logo"
        self.assertTrue(self.openrazer._has_zone_capability(rdevice, zone, "reactive"), "Capability incorrectly reported")

    def test_capability_scroll(self):
        rdevice = self.get_rdevice("Razer Taipan")
        zone = OpenRazerBackend.DeviceItem.Zone()
        zone.zone_id = "scroll"
        self.assertTrue(self.openrazer._has_zone_capability(rdevice, zone, "on"), "Capability incorrectly reported")

    def test_capability_backlight(self):
        rdevice = self.get_rdevice("Razer Naga 2012")
        zone = OpenRazerBackend.DeviceItem.Zone()
        zone.zone_id = "backlight"
        self.assertTrue(self.openrazer._has_zone_capability(rdevice, zone, "on"), "Capability incorrectly reported")

    def test_capability_left(self):
        rdevice = self.get_rdevice("Razer Mamba Elite")
        zone = OpenRazerBackend.DeviceItem.Zone()
        zone.zone_id = "left"
        self.assertTrue(self.openrazer._has_zone_capability(rdevice, zone, "spectrum"), "Capability incorrectly reported")

    def test_capability_right(self):
        rdevice = self.get_rdevice("Razer Naga Left-Handed Edition 2020")
        zone = OpenRazerBackend.DeviceItem.Zone()
        zone.zone_id = "right"
        self.assertTrue(self.openrazer._has_zone_capability(rdevice, zone, "static"), "Capability incorrectly reported")

    def test_capability_charging(self):
        rdevice = self.get_rdevice("Razer Charging Pad Chroma")
        zone = OpenRazerBackend.DeviceItem.Zone()
        zone.zone_id = "charging"
        self.assertTrue(self.openrazer._has_zone_capability(rdevice, zone, "breath_random"), "Capability incorrectly reported")

    def test_capability_fast_charging(self):
        rdevice = self.get_rdevice("Razer Charging Pad Chroma")
        zone = OpenRazerBackend.DeviceItem.Zone()
        zone.zone_id = "fast_charging"
        self.assertTrue(self.openrazer._has_zone_capability(rdevice, zone, "brightness"), "Capability incorrectly reported")

    def test_capability_fully_charged(self):
        rdevice = self.get_rdevice("Razer Charging Pad Chroma")
        zone = OpenRazerBackend.DeviceItem.Zone()
        zone.zone_id = "fully_charged"
        self.assertTrue(self.openrazer._has_zone_capability(rdevice, zone, "wave"), "Capability incorrectly reported")

    def test_option_colours_persist(self):
        device = self.get_device("Razer Mamba Tournament Edition")
        static = self.get_option(device, "static", "backlight")
        static.colours[0] = "#123456"
        static.apply()
        static.colours[0] = "#000000"
        device.refresh()
        self.assertTrue(static.colours[0] == "#123456", "Persistence did not read/write as expected.")

    def test_option_brightness_backlight(self):
        device = self.get_device("Razer Mamba Tournament Edition")
        brightness = self.get_option(device, "brightness", "backlight")
        brightness.apply(24)
        brightness.refresh()
        self.assertTrue(brightness.value == 24, "Could not set 'variable' brightness correctly")

    def test_option_brightness_logo(self):
        device = self.get_device("Razer Mamba Elite")
        brightness = self.get_option(device, "brightness", "logo")
        brightness.apply(48)
        brightness.refresh()
        self.assertTrue(brightness.value == 48, "Could not set 'variable' brightness correctly")

    def test_option_brightness_right(self):
        device = self.get_device("Razer Lancehead (Wired)")
        brightness = self.get_option(device, "brightness", "right")
        brightness.apply(72)
        brightness.refresh()
        self.assertTrue(brightness.value == 72, "Could not set 'variable' brightness correctly")

    def test_option_on_logo(self):
        self.skipTest("FIXME: Buggy test")
        device = self.get_device("Razer Taipan")
        option = self.get_option(device, "on", "logo")
        option.apply(False)
        option.refresh()
        self.assertTrue(option.active == False, "Could not unset 'on' logo effect correctly")

    def test_option_on_scroll(self):
        self.skipTest("FIXME: Buggy test")
        device = self.get_device("Razer Taipan")
        option = self.get_option(device, "on", "scroll")
        option.apply(True)
        option.refresh()
        self.assertTrue(option.active == True, "Could not set 'on' scroll effect correctly")

    def test_option_effect_change(self):
        device = self.get_device("Razer BlackWidow Chroma")
        none = self.get_option(device, "none", "main")
        static = self.get_option(device, "static", "main")
        none.apply()
        static.apply()
        none.refresh()
        device.refresh()
        self.assertTrue(none.active == False and static.active == True, "Could not set effect active state")

    def test_option_effect_none(self):
        device = self.get_device("Razer BlackWidow Lite")
        effect = self.get_option(device, "none", "main")
        effect.apply()

    def test_option_effect_spectrum(self):
        device = self.get_device("Razer Blade Pro (2019)")
        effect = self.get_option(device, "spectrum", "main")
        effect.apply()

    def test_option_effect_wave(self):
        device = self.get_device("Razer Blade (Late 2016)")
        effect = self.get_option(device, "wave", "main")
        effect.apply(effect.parameters[0].data)

    def test_option_effect_wave_label_mouse(self):
        device = self.get_device("Razer Mamba (Wireless)")
        effect = self.get_option(device, "wave", "backlight")
        self.assertTrue(effect.parameters[1].label == "Up", "Unexpected wave label for mouse")

    def test_option_effect_wave_label_mousemat(self):
        device = self.get_device("Razer Firefly V2")
        effect = self.get_option(device, "wave", "main")
        self.assertTrue(effect.parameters[1].label == "Clockwise", "Unexpected wave label for mousemat")

    def test_option_effect_wheel(self):
        device = self.get_device("Razer BlackWidow V4 Pro")
        effect = self.get_option(device, "wheel", "main")
        effect.apply(effect.parameters[0].data)

    def test_option_effect_ripple(self):
        device = self.get_device("Razer BlackWidow Chroma")
        effect = self.get_option(device, "ripple", "main")
        effect.apply(effect.parameters[0].data)

    def test_option_effect_reactive(self):
        device = self.get_device("Razer BlackWidow V3")
        effect = self.get_option(device, "reactive", "main")
        effect.apply(effect.parameters[0].data)

    def test_option_effect_static(self):
        device = self.get_device("Razer BlackWidow Chroma")
        effect = self.get_option(device, "static", "main")
        effect.apply()

    def test_option_effect_breath(self):
        device = self.get_device("Razer BlackWidow 2019")
        effect = self.get_option(device, "breath", "main")
        effect.apply(effect.parameters[0].data)

    def test_option_effect_breath_triple(self):
        device = self.get_device("Razer Kraken Ultimate")
        effect = self.get_option(device, "breath", "main")
        # Expects: [single, dual, triple]
        effect.apply(effect.parameters[2].data)

    def test_option_effect_starlight(self):
        device = self.get_device("Razer BlackWidow X Ultimate")
        effect = self.get_option(device, "starlight", "main")
        effect.apply(effect.parameters[-1].data)

    def test_workaround_bw2013_pulsate(self):
        device = self.get_device("Razer Deathstalker Expert")
        effect = self.get_option(device, "pulsate", "main")
        effect.apply()

    def test_workaround_bw2013_static(self):
        device = self.get_device("Razer BlackWidow Ultimate 2013")
        effect = self.get_option(device, "static", "main")
        effect.apply()

    def test_poll_rate_set(self):
        device = self.get_device("Razer Taipan")
        poll_rate = self.get_option(device, "poll_rate", "main")
        poll_rate.apply(500)

    def test_poll_rate_set_hyper(self):
        device = self.get_device("Razer Viper 8KHz")
        poll_rate = self.get_option(device, "poll_rate", "main")
        poll_rate.apply(8000)

    def test_game_mode_set(self):
        device = self.get_device("Razer BlackWidow Chroma")
        game_mode = self.get_option(device, "game_mode", "main")
        game_mode.apply(True)

    def test_battery_idle_time_set_only(self):
        device = self.get_device("Razer Ouroboros")
        idle_time = self.get_option(device, "idle_time", "main")
        idle_time.apply(10)
        # 'get' is powered by fallback persistence
        device.refresh()
        idle_time.refresh()
        self.assertEqual(idle_time.value, 10, "Could not set or get idle_time (device = set only)")

    def test_battery_low_power_set_only(self):
        device = self.get_device("Razer Ouroboros")
        low_bat_thres = self.get_option(device, "low_battery_threshold", "main")
        low_bat_thres.apply(50)
        # 'get' is powered by fallback persistence
        device.refresh()
        low_bat_thres.refresh()
        self.assertEqual(low_bat_thres.value, 50, "Could not set or get low_battery_threshold (device = set only)")

    def test_battery_idle_time_set_get(self):
        device = self.get_device("Razer Basilisk X HyperSpeed")
        idle_time = self.get_option(device, "idle_time", "main")
        idle_time.apply(10)
        # 'get' is retrieved from device
        device.refresh()
        idle_time.refresh()
        self.assertEqual(idle_time.value, 10, "Could not set or get idle_time (device = get/set)")

    def test_battery_low_power_set_get(self):
        device = self.get_device("Razer Basilisk X HyperSpeed")
        low_bat_thres = self.get_option(device, "low_battery_threshold", "main")
        low_bat_thres.apply(50)
        # 'get' is retrieved from device
        device.refresh()
        low_bat_thres.refresh()
        self.assertEqual(low_bat_thres.value, 50, "Could not set or get low_battery_threshold (device = get/set)")

    def test_scroll_mode_set_get(self):
        device = self.get_device("Razer Basilisk V3")
        scroll_mode = self.get_option(device, "scroll_mode", "main")
        scroll_mode.apply(1)
        scroll_mode.refresh()

    def test_scroll_accel_set_get(self):
        device = self.get_device("Razer Basilisk V3")
        scroll_accel = self.get_option(device, "scroll_accel", "main")
        scroll_accel.apply(1)
        scroll_accel.refresh()

    def test_smart_reel_set_get(self):
        device = self.get_device("Razer Basilisk V3")
        smart_reel = self.get_option(device, "scroll_smart_reel", "main")
        smart_reel.apply(True)
        smart_reel.refresh()

    def test_all_options_no_duplicates(self):
        for device in self.openrazer.get_devices():
            all_options = []
            for zone in device.zones:
                all_options += [zone.zone_id + "-" + option.uid for option in zone.options]

            filtered_options = set(all_options)
            if len(filtered_options) < len(all_options):
                raise KeyError("Duplicate option UIDs for " + device.name + \
                    f"! {len(all_options)} UIDs but only {len(filtered_options)} unique: {str(all_options)}")

    def test_all_devices_refresh(self):
        for device in self.openrazer.get_devices():
            device.refresh()

    def test_all_options_refresh(self):
        for device in self.openrazer.get_devices():
            for zone in device.zones:
                for option in zone.options:
                    option.refresh()

    def test_all_options_apply(self):
        for device in self.openrazer.get_devices():
            device.refresh()
            for zone in device.zones:
                for option in zone.options:
                    if isinstance(option, Backend.EffectOption):
                        if option.parameters:
                            option.apply(random.choices(option.parameters)[0].data)
                        else:
                            option.apply()
                    elif isinstance(option, Backend.ToggleOption):
                        option.apply(False)
                        option.apply(True)
                    elif isinstance(option, Backend.SliderOption):
                        option.apply(random.choices(range(option.min, option.max, option.step))[0])
                    elif isinstance(option, Backend.MultipleChoiceOption):
                        option.apply(random.choices(option.parameters)[0].data)
                    option.refresh()

    def test_persistence_colour_bytes(self):
        rdevice = self.get_rdevice("Razer Mamba Tournament Edition")
        persistence = OpenRazerPersistence(rdevice.fx.misc.backlight)
        persistence._convert_colour_bytes(rdevice.fx.misc.backlight)

    def test_persistence_refresh(self):
        rdevice = self.get_rdevice("Razer Mamba Tournament Edition")
        persistence = OpenRazerPersistence(rdevice.fx.misc.backlight)
        persistence.refresh()

    def test_persistence_fallback_read(self):
        rdevice = self.get_rdevice("Razer Mamba Tournament Edition")
        persistence = OpenRazerPersistenceFallback("main", rdevice.serial, self.openrazer.persistence_fallback_path)
        persistence.refresh()

    def test_persistence_fallback_write(self):
        rdevice = self.get_rdevice("Razer Mamba Tournament Edition")
        persistence = OpenRazerPersistenceFallback("main", rdevice.serial, self.openrazer.persistence_fallback_path)
        persistence.save("effect", "static")


if __name__ == '__main__':
    unittest.main()



# Polychromatic is licensed under the GPLv3.
# Copyright (C) 2020-2024 Luke Horwell <code@horwell.me>
"""
This module controls the 'Devices' tab of the Controller GUI.
"""

import json
from polychromatic.procpid import DeviceSoftwareState
from .. import bulkapply
from .. import common
from .. import effects
from .. import locales
from .. import preferences as pref
from .. import middleman
from . import shared
from ..backends._backend import Backend as Backend
from ..qt.flowlayout import FlowLayout as QFlowLayout

import os
import subprocess
import time
import shutil
import webbrowser

from PyQt5.QtCore import Qt, QSize, QMargins, QThread
from PyQt5.QtGui import QIcon, QPixmap, QFont
from PyQt5.QtWidgets import QWidget, QDialogButtonBox, QGroupBox, QGridLayout, \
                            QPushButton, QToolButton, QMessageBox, QListWidget, \
                            QTreeWidget, QTreeWidgetItem, QLabel, QComboBox, \
                            QSpacerItem, QSizePolicy, QSlider, QCheckBox, \
                            QButtonGroup, QRadioButton, QDialog, QTableWidget, \
                            QTableWidgetItem, QAction, QHBoxLayout

# Error codes
ERROR_NO_DEVICE = 0
ERROR_BACKEND_IMPORT = 1
ERROR_NO_BACKEND = 2


class DevicesTab(shared.TabData):
    """
    Allows the user to quickly change the existing state of the device right now.
    """
    def __init__(self, appdata):
        super().__init__(appdata)
        self.special_controls = SpecialControls(appdata)

        # Session
        self.current_device = None
        self.load_thread = None

        # UI Elements
        self.Contents = self.main_window.findChild(QWidget, "DeviceContents")
        self.SidebarTree = self.main_window.findChild(QTreeWidget, "DeviceSidebarTree")
        self.SidebarTree.itemClicked.connect(self._sidebar_changed)

        # Useful complimentary software
        self.input_remapper = shutil.which("input-remapper-gtk")

        # Avoid garbage collection cleaning up invisible controls
        self.btn_grps = {}

    def set_tab(self):
        """
        Device tab opened. Populate the device and task lists, and open the
        properties for the first device (if applicable)
        """
        self.set_title(self._("Devices"))

        # Open all 'sidebar' tree branches
        self.SidebarTree.expandAll()
        self.SidebarTree.setEnabled(True)

        # Populate sidebar
        tasks_branch = self.SidebarTree.invisibleRootItem().child(0)
        devices_branch = self.SidebarTree.invisibleRootItem().child(1)

        # Give IDs for fixed items
        tasks_branch.section = None
        devices_branch.section = None
        tasks_branch.child(0).section = "apply-to-all"

        # Backends are still initialising
        if not self.appdata.ready:
            self._open_loading()
            return

        # Recache the device list
        self.SidebarTree.parent().show()
        device_list = self.middleman.get_devices()
        unknown_device_list = self.middleman.get_unsupported_devices()

        # Clear device entries
        devices_branch.takeChildren()

        for device in device_list:
            item = QTreeWidgetItem()
            item.setText(0, device.name)
            item.setIcon(0, QIcon(device.form_factor["icon"]))
            item.section = "device"
            item.device_item = device
            devices_branch.addChild(item)

        devices_branch.sortChildren(0, Qt.AscendingOrder)

        # For unknown devices, check support status using local index
        device_index = json.loads(open(f"{self.paths.data_dir}/devices/openrazer.json", "r").read())
        supported_pids = device_index.keys()

        for device in unknown_device_list:
            vidpid = f"{device.vid}:{device.pid}"

            if vidpid in supported_pids:
                device.name = device_index[vidpid].get("name", device.name)
                device.supported = True

            item = QTreeWidgetItem()
            item.setText(0, device.name)
            item.setIcon(0, QIcon(device.form_factor["icon"]))
            item.section = "device-unknown"
            item.device_item = device
            devices_branch.addChild(item)

        # Only show tasks when there are multiple devices present
        if len(device_list) > 1:
            tasks_branch.setHidden(False)
        else:
            tasks_branch.setHidden(True)

        # Backends loaded, but no usable devices
        if len(device_list) == 0 and len(unknown_device_list) > 0:
            first_item = devices_branch.child(0)
            first_item.setSelected(True)
            self.open_unknown_device(first_item.device_item)

        # All backends failed to load
        elif len(self.middleman.backends) == 0 and len(self.middleman.import_errors) > 0:
            self._open_no_backend_found(ERROR_BACKEND_IMPORT)

        # No backends are installed
        elif len(self.middleman.backends) == 0 and len(self.middleman.not_installed) > 0:
            self._open_no_backend_found(ERROR_NO_BACKEND)

        # Backends present, but no devices listed
        elif len(device_list) == 0:
            self._open_no_backend_found(ERROR_NO_DEVICE)

        # Open the first device initially
        if len(device_list) > 0:
            first_item = devices_branch.child(0)
            first_item.setSelected(True)
            self.open_device(first_item.device_item)

    def _sidebar_changed(self, item):
        """
        User chooses an item on the sidebar. The "section" variable is appended
        directly into the QTreeWidgetItem.
        """
        if item.section == "apply-to-all":
            self.open_apply_to_all()
        elif item.section == "device-unknown":
            self.open_unknown_device(item.device_item)
        elif item.section == "device":
            self.open_device(item.device_item)

    def _get_device_summary_widget(self, device):
        """
        Generate the heading at the top of the page that shows the name, image
        and current state of the selected device.
        """
        badges = []
        # TODO: Need new state class
        #current_sw_effect = device["state"]["effect"]
        #current_preset = device["state"]["preset"]

        # Go through each zone and find out if things are the same
        effects_used = []
        effect_name = ""
        effect_icon = ""
        brightnesses = []

        for zone in device.zones:
            for option in zone.options:
                if isinstance(option, Backend.SliderOption) and option.uid == "brightness":
                    brightnesses.append(option.value)
                if isinstance(option, Backend.EffectOption) and option.active:
                    effects_used.append(option.uid)
                    effect_name = option.label
                    effect_icon = option.icon

        state = DeviceSoftwareState(device.serial)
        sw_effect = state.get_effect()
        if sw_effect:
            badges.append({
                "label": sw_effect.get("name"),
                "icon": sw_effect.get("icon"),
            })
        elif effects_used and all(i == effects_used[0] for i in effects_used):
            badges.append({
                "label": effect_name,
                "icon": effect_icon,
            })
        elif effects_used:
            badges.append({
                "label": self._("(Varies)"),
                "icon": common.get_icon("general", "effects"),
            })

        if brightnesses and all(i == brightnesses[0] for i in brightnesses):
            value = brightnesses[0]
            if value >= 40:
                icon = common.get_icon("params", "100")
            elif value >= 1:
                icon = common.get_icon("params", "50")
            else:
                icon = common.get_icon("params", "0")

            badges.append({
                "label": str(value) + "%",
                "icon": icon,
            })
        elif brightnesses:
            badges.append({
                "label": self._("(Varies)"),
                "icon": common.get_icon("options", "brightness"),
            })

        if device.dpi:
            if device.dpi.x == device.dpi.y:
                label = f"{device.dpi.x} DPI"
            else:
                label = f"{device.dpi.x}, {device.dpi.y} DPI"

            badges.append({
                "label": label,
                "icon": common.get_icon("general", "dpi"),
            })

        if device.battery:
            label = f"{str(device.battery.percentage)}%"
            if device.battery.percentage >= 90:
                icon = "battery-100"
            elif device.battery.percentage >= 70:
                icon = "battery-75"
            elif device.battery.percentage >= 40:
                icon = "battery-50"
            elif device.battery.percentage >= 15:
                icon = "battery-25"
            else:
                icon = "battery-0"

            if device.battery.is_charging:
                icon = "battery-charging"
                label += " (charging)"

            badges.append({
                "label": label,
                "icon": common.get_icon("general", icon),
            })

        # TODO: Add additional preset state

        if device.real_image:
            real_image = shared.get_real_device_image(device.real_image)
        else:
            real_image = common.get_icon("devices", "noimage")

        # TODO: Move buttons to toolbar
        buttons = [
            {
                "id": "device-info",
                "icon": None,
                "label": self._("Device Info"),
                "disabled": False,
                "action": self.show_device_info
            }
        ]

        return self.widgets.create_summary_widget(real_image, device.name, badges, buttons)

    def open_device(self, device):
        """
        Show details and controls for changing the current hardware state
        of the specified device.
        """
        self.current_device = device

        self.set_cursor_busy()
        layout = self.Contents.layout()
        layout.setContentsMargins(0, 0, 0, 0)
        shared.clear_layout(layout)

        try:
            device.refresh()
        except Exception as e:
            # State may have changed, reload the device tab.
            # TODO: Show "device changes detected" in status bar. Needs global function.
            self.middleman.invalidate_cache()
            return self.set_tab()

        layout.addWidget(self._get_device_summary_widget(device))

        for zone in device.zones:
            widgets = []
            first_zone = True if device.zones[0] == zone else False

            # Effects are all 'collapsed' into one row. When encountering the first
            # effect, remember the index to insert here later.
            has_effects = False
            effects_index = 0

            for index, option in enumerate(zone.options):
                if isinstance(option, Backend.EffectOption):
                    if not has_effects:
                        has_effects = True
                        effects_index = index
                    continue

                widgets.append(self._create_row_control(option))

            # Controls for effects, its parameters and colours
            def _get_effect_options(zone, force_inactive):
                options = []
                for option in zone.options:
                    # Override active flag when software effect is running.
                    if force_inactive:
                        option.active = False

                    if isinstance(option, Backend.EffectOption):
                        options.append(option)

                return options

            state = DeviceSoftwareState(device.serial)
            sw_effect = state.get_effect()
            effect_options = _get_effect_options(zone, True if sw_effect else False)
            active_effect = self.middleman.get_active_effect(zone)
            effect_row = self._create_effect_controls(zone, effect_options)
            if effect_row:
                widgets.insert(effects_index, effect_row)
                effects_index += 1

            # Show parameters and colours for selected effect (if applicable)
            if active_effect:
                param_row = self._create_effect_parameter_controls(device, zone, active_effect)
                if param_row:
                    widgets.insert(effects_index, param_row)
                    effects_index += 1

                colours_required = self.middleman.get_active_colours_required(active_effect)
                if colours_required > 0:
                    for pos in range(0, colours_required):
                        widgets.insert(effects_index, self._create_colour_control(device, zone, pos, active_effect.colours))
                        effects_index += 1

            # General device controls
            if first_zone:
                # -- Mouse DPI
                if device.dpi:
                    widgets.append(self.special_controls.create_dpi_control(device))

                # TODO: Button to set compatible software effects - or as "Custom" effect?

                # -- Mouse Acceleration
                # TODO: Move to middleman for all interfaces?
                if device.form_factor == "mouse":
                    widgets.append(self.special_controls.create_mouse_accel_control())

                # -- Information about OpenRazer daemon's built-in macro feature (deprecated)
                if device.has_macro_keys:
                   widgets.append(self.special_controls.create_openrazer_macro_control())

                # -- Information about programmable keys
                if device.has_programmable_keys and self.input_remapper:
                    widgets.append(self.special_controls.create_programmable_keys_control(self.input_remapper))

            # Before adding to the main layout, use group boxes if there's multiple zones
            if len(device.zones) > 1:
                group = self.widgets.create_group_widget(zone.label)
                for widget in widgets:
                    group.layout().addWidget(widget)
                layout.addWidget(group)
            else:
                for widget in widgets:
                    layout.addWidget(widget)

        layout.addStretch()
        self.set_cursor_normal()

    def reload_device(self):
        """
        Reloads the current device page.
        """
        self.open_device(self.current_device)

    def _create_row_control(self, option):
        """
        Returns a list of widgets for the specified option.
        """
        if isinstance(option, Backend.SliderOption):
            return self.widgets.create_row_widget(option.label, self._create_control_slider(option))

        elif isinstance(option, Backend.ToggleOption):
            return self.widgets.create_row_widget(option.label, self._create_control_toggle(option))

        elif isinstance(option, Backend.MultipleChoiceOption):
            return self.widgets.create_row_widget(option.label, self._create_control_select(option))

    def _create_control_slider(self, option):
        """
        Returns a list of controls that make up a slider for changing a variable option.
        """
        slider = QSlider(Qt.Horizontal)
        slider.setValue(option.value)
        slider.setMinimum(option.min)
        slider.setMaximum(option.max)
        slider.setSingleStep(option.step)
        slider.setPageStep(option.step * 2)
        slider.setMaximumWidth(150)

        # BUG: Qt: Ticks don't appear with stylesheet
        slider.setTickPosition(QSlider.TicksBelow)
        slider.setTickInterval(int(option.max / 10))

        label = QLabel()
        suffix = option.suffix if option.value == 1 else option.suffix_plural
        label.setText(str(option.value) + suffix)

        # Change label while sliding
        def _slider_moved(value):
            suffix = option.suffix if value == 1 else option.suffix_plural
            label.setText(str(value) + suffix)
        slider.sliderMoved.connect(_slider_moved)
        slider.valueChanged.connect(_slider_moved)

        # Send request once dropped
        def _slider_dropped():
            self.dbg.stdout(f"{self.current_device.name}: Applying option {option.uid} with value: {str(slider.value())}", self.dbg.action, 1)
            try:
                option.apply(slider.value())
            except Exception as e:
                self._catch_command_error(self.current_device, e)

        slider.sliderReleased.connect(_slider_dropped)
        slider.valueChanged.connect(_slider_dropped)

        return [slider, label]

    def _create_control_toggle(self, option):
        """
        Returns a list of controls that make up a toggle for a binary choices.
        """
        checkbox = QCheckBox(option.label_toggle)

        if option.active:
            checkbox.setChecked(option.active)

        def _state_changed():
            onoff = "on" if checkbox.isChecked() else "off"
            self.dbg.stdout(f"{self.current_device.name}: Turning {onoff} option {option.uid}", self.dbg.action, 1)
            try:
                option.apply(checkbox.isChecked())
            except Exception as e:
                self._catch_command_error(self.current_device, e)

        checkbox.stateChanged.connect(_state_changed)
        return [checkbox]

    def _create_control_select(self, option):
        """
        Returns a list of controls that make up a multiple choice dropdown.
        """
        params = option.parameters
        selected_index = 0
        combo = QComboBox()
        combo.setIconSize(QSize(16, 16))

        for index, param in enumerate(params):
            combo.addItem(param.label)
            if param.icon:
                combo.setItemIcon(combo.count() - 1, QIcon(param.icon))
            if param.active:
                selected_index = index

        combo.setCurrentIndex(selected_index)

        def _current_index_changed(index):
            param = params[index]
            self.dbg.stdout(f"{self.current_device.name}: Setting option {option.uid} to {param.data}", self.dbg.action, 1)
            try:
                option.apply(param.data)
            except Exception as e:
                self._catch_command_error(self.current_device, e)

        combo.currentIndexChanged.connect(_current_index_changed)
        return [combo]

    def _create_effect_controls(self, zone, options):
        """
        Return a row widget containing the specified options. These are grouped
        together and will be presented as larger buttons.
        """
        widgets = []
        self.btn_grps[zone] = QButtonGroup()

        for option in options:
            button = QToolButton()
            button.setText(option.label)
            button.setCheckable(True)
            button.setIconSize(QSize(40, 40))
            button.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)
            if option.icon:
                button.setIcon(QIcon(option.icon))
            button.setMinimumHeight(70)
            button.setMinimumWidth(105)
            button.option = option
            self.btn_grps[zone].addButton(button)

            if option.active:
                button.setChecked(True)

            widgets.append(button)

        def _clicked_effect_button(button):
            option = button.option
            param = self.middleman.get_default_parameter(option)
            self.middleman.stop_software_effect(self.current_device.serial)

            try:
                if param:
                    self.dbg.stdout(f"{self.current_device.name}: Setting effect {option.uid} (with parameter {str(param.data)}')", self.dbg.action, 1)
                    option.apply(param.data)
                else:
                    self.dbg.stdout(f"{self.current_device.name}: Setting effect {option.uid} (no parameters)", self.dbg.action, 1)
                    option.apply()
            except Exception as e:
                self._catch_command_error(self.current_device, e)

            self.reload_device()

        self.btn_grps[zone].buttonClicked.connect(_clicked_effect_button)

        if not widgets:
            return None

        return self.widgets.create_row_widget(self._("Effect"), widgets, wrap=True)

    def _create_effect_parameter_controls(self, device, zone, option):
        """
        Returns a list of row widgets that change the active effect's parameters.
        """
        if not option.parameters:
            return None

        def _clicked_param_button():
            for radio in self.btn_grps["radio_param_" + zone.zone_id]:
                if not radio.isChecked():
                    continue
                self.dbg.stdout(f"{device.name}: Setting parameter for '{option.uid}' to '{radio.param.data}'", self.dbg.action, 1)
                try:
                    option.apply(radio.param.data)
                except Exception as e:
                    self._catch_command_error(self.current_device, e)
                self.reload_device()

        def _make_button(param):
            radio = QRadioButton()
            radio.setText(param.label)
            radio.clicked.connect(_clicked_param_button)
            radio.option = option
            radio.param = param

            if param.active:
                radio.setChecked(True)

            if len(option.parameters) == 1 and param.active:
                radio.setDisabled(True)

            if param.icon:
                radio.setIcon(QIcon(param.icon))
                radio.setIconSize(QSize(22, 22))
            return radio

        widgets = []
        for param in option.parameters:
            widgets.append(_make_button(param))

        self.btn_grps["radio_param_" + zone.zone_id] = widgets
        return self.widgets.create_row_widget(self._("Effect Mode"), widgets, vertical=True)

    def _create_colour_control(self, device, zone, pos, colours):
        """
        Returns a row widget to set a colour at a specified position for the active option/parameter.
        Position is 0-based, so 0 = primary, 1 = secondary, etc
        """
        pretty_labels = {
            0: self._("Primary Colour"),
            1: self._("Secondary Colour"),
            2: self._("Tertiary Colour"),
        }

        try:
            label = pretty_labels[pos]
        except KeyError:
            label = self._("Colour 4").replace("4", str(pos))

        def _set_new_colour(new_hex, data):
            self.dbg.stdout(f"{self.current_device.name}: Setting {label.lower()} to {new_hex} for active option in zone {zone.zone_id}", self.dbg.action, 1)
            try:
                self.middleman.set_colour_for_active_effect_zone(zone, new_hex, pos)
            except Exception as e:
                self._catch_command_error(self.current_device, e)

        _set_data = {"zone": zone, "colour_no": pos}

        colour_widget = self.widgets.create_colour_control(colours[pos], _set_new_colour, _set_data, self._("Change []").replace("[]", label), device.monochromatic)
        return self.widgets.create_row_widget(label, [colour_widget])

    def _catch_command_error(self, device=Backend.DeviceItem, err=""):
        """
        Shows a error message when a command sent to a device was unsuccessful.
        """
        backend_name = device.backend.name
        exception = common.get_exception_as_string(err)
        my_fault = common.is_exception_fault_by_app(err)

        self.dbg.stdout(exception, self.dbg.error)

        if my_fault:
            self.widgets.open_dialog(self.widgets.dialog_error,
                                     self._("Request Failed"),
                                     self._("There's a fault with Polychromatic's integration with [OpenRazer] for this device or option.").replace("[OpenRazer]", backend_name),
                                     self._("To help fix this, please check Polychromatic's issue tracker and create a new issue if a similar one does not exist.").replace("[OpenRazer]", backend_name),
                                     details=exception)
        else:
            self.widgets.open_dialog(self.widgets.dialog_error,
                                     self._("Communication Error"),
                                     self._("An error occurred outside of Polychromatic's control when this command was sent to [OpenRazer].").replace("[OpenRazer]", backend_name),
                                     self._("Try restarting [OpenRazer] and Polychromatic and give it another go. If this error appears again, it might need fixing in [OpenRazer].").replace("[OpenRazer]", backend_name),
                                     details=exception)

    def _open_loading(self):
        """
        Show a skeleton loading screen while backends finish loading.

        A thread will spawn to refresh the devices tab (if open) when the
        middleman has fully populated.
        """
        if self.load_thread:
            return

        # Sidebar
        self.SidebarTree.setEnabled(False)
        tasks_branch = self.SidebarTree.invisibleRootItem().child(0)
        devices_branch = self.SidebarTree.invisibleRootItem().child(1)
        tasks_branch.setHidden(True)

        # Skeleton devices
        device_icon = QIcon()
        device_icon.addFile(common.get_icon("general", "placeholder"), mode=QIcon.Disabled)
        skel_strings = [
            "█████████████",
            "████████████",
            "█████████",
            "███████████",
            "███████",
        ]
        for i in range(0, 5):
            skel = QTreeWidgetItem()
            skel.setText(0, skel_strings[i])
            skel.setIcon(0, device_icon)
            skel.setDisabled(True)
            devices_branch.addChild(skel)

        # Skeleton summary
        def dummy():
            pass

        buttons = []
        indicators = [
            {
                "icon": common.get_icon("general", "placeholder"),
                "label": "██████"
            },
            {
                "icon": common.get_icon("general", "placeholder"),
                "label": "███"
            }
        ]
        summary = self.widgets.create_summary_widget(common.get_icon("devices", "noimage"), "████████████", indicators, buttons)

        # Skeleton device controls (brightness & effect)
        skels_text = summary.findChildren(QLabel)
        skels_bg = []

        dummy_slider = QLabel("________________________")
        dummy_slider.setMaximumWidth(180)
        dummy_slider.setAlignment(Qt.AlignLeft)
        skels_bg.append(dummy_slider)

        dummy_buttons = []
        for i in range(0, 5):
            dummy = QLabel()
            dummy.setMinimumHeight(70)
            dummy.setMinimumWidth(70)
            dummy.setMaximumHeight(70)
            dummy.setMaximumWidth(70)
            dummy.setAlignment(Qt.AlignLeft)
            dummy_buttons.append(dummy)
            skels_bg.append(dummy)

        row1 = self.widgets.create_row_widget("███████", [dummy_slider], wrap=True)
        row2 = self.widgets.create_row_widget("████", dummy_buttons, wrap=True)

        label1 = row1.findChild(QLabel)
        label2 = row2.findChild(QLabel)

        skels_text.append(label1)
        skels_text.append(label2)

        # Apply skeleton styling
        for skel in skels_bg:
            skel.setStyleSheet("QLabel { background-color: #202020; color: #202020; margin: 4px 0; }")
        for skel in skels_text:
            skel.setStyleSheet("QLabel { color: #202020; }")

        # Append widgets
        layout = self.Contents.layout()
        for widget in [summary, row1, row2]:
            layout.addWidget(widget)
        layout.addStretch()

        class WaitForBackendThread(QThread):
            @staticmethod
            def run():
                while not self.appdata.ready:
                    time.sleep(0.05)

                refresh = self.main_window.findChild(QAction, "actionRefreshTab")
                refresh.trigger()

        self.load_thread = WaitForBackendThread()
        self.load_thread.start()
        self.set_cursor_busy()

    def _open_no_backend_found(self, message_id):
        """
        No backends are present. Hide the sidebar and show a full screen message.

        Params:
            message_id      (int)   An self.err_* integer.
        """
        self.SidebarTree.parent().hide()

        layout = self.Contents.layout()
        layout.setContentsMargins(0, 0, 0, 0)
        shared.clear_layout(layout)

        graphic = {
            ERROR_NO_DEVICE: common.get_icon("empty", "nodevice"),
            ERROR_BACKEND_IMPORT: common.get_icon("empty", "nobackend"),
            ERROR_NO_BACKEND: common.get_icon("empty", "nobackend")
        }

        title = {
            ERROR_NO_DEVICE: self._("No devices connected"),
            ERROR_BACKEND_IMPORT: self._("No backends loaded"),
            ERROR_NO_BACKEND: self._("No backends installed")
        }

        subtitle = {
            ERROR_NO_DEVICE: self._("Plug in a compatible device to control its lighting effects and features"),
            ERROR_BACKEND_IMPORT: self._("Consult the logs and troubleshooter for hints on fixing this problem"),
            ERROR_NO_BACKEND: self._("Install a compatible backend to configure lighting effects")
        }

        buttons = {
            ERROR_NO_DEVICE: [
                {
                    "label": self._("Troubleshoot"),
                    "icon_folder": "general",
                    "icon_name": "preferences",
                    "action": self._start_troubleshooter
                }
            ],
            ERROR_NO_BACKEND: [
                {
                    "label": self._("Online Help"),
                    "icon_folder": "general",
                    "icon_name": "external",
                    "action": self._open_openrazer_help_unrecognised
                },
                {
                    "label": self._("Troubleshoot"),
                    "icon_folder": "general",
                    "icon_name": "preferences",
                    "action": self._start_troubleshooter
                }
            ]
        }
        buttons[ERROR_BACKEND_IMPORT] = buttons[ERROR_NO_BACKEND].copy()
        buttons[ERROR_BACKEND_IMPORT].append({
            "label": self._("Show Error Details"),
            "icon_folder": "general",
            "icon_name": "warning",
            "action": self._open_backend_exception
        })

        self.widgets.populate_empty_state(layout, graphic[message_id], title[message_id], subtitle[message_id], buttons[message_id])

    def _open_online_help(self):
        """
        Opens the documentation online to learn more about Polychromatic's device
        functionality.
        """
        self.appdata.menubar.online_help()

    def _start_troubleshooter(self):
        """
        Starts the specified troubleshooter; the troubleshooter for the only
        backend available, or shows a prompt for the user to choose the backend
        to troubleshoot.
        """
        # TODO: Prompt not implemented as only one backend (OpenRazer)
        self.appdata.menubar.openrazer.troubleshoot()

    def _open_backend_exception(self):
        """
        Opens a dialog showing the details of an exception for one of the backends.
        """
        dbg = self.dbg
        _ = self._

        for backend_id in self.middleman.import_errors.keys():
            backend_name = "openrazer"
            exception = self.middleman.import_errors[backend_id].strip()
            dbg.stdout("\n{0}\n------------------------------\n{1}\n".format(backend_name, exception), dbg.error)
            self.widgets.open_dialog(self.widgets.dialog_generic,
                                    _("Backend Error: []").replace("[]", backend_name),
                                    _("An error occurred trying to load []. The error below may provide a clue to what happened.").replace("[]", backend_name),
                                    info_text=_("The last line of the exception was:") + "\n" + exception.split("\n")[-1],
                                    details=exception)

    def _open_openrazer_help_unrecognised(self):
        self.appdata.menubar._prompt_on_locale_change(self._("Online Help"))
        webbrowser.open("https://docs.polychromatic.app/openrazer/#my-device-is-showing-up-as-unrecognised")

    def _open_openrazer_help_unsupported(self):
        self.appdata.menubar._prompt_on_locale_change(self._("Online Help"))
        webbrowser.open("https://docs.polychromatic.app/openrazer/#my-device-isnt-listed-as-supported-what-do-i-do")

    def open_unknown_device(self, device):
        """
        Show guidance on a device that could be controlled, but isn't possible right now.
        """
        layout = self.Contents.layout()
        layout.setContentsMargins(0, 0, 0, 0)
        shared.clear_layout(layout)

        backend_name = "OpenRazer"

        def _restart_backend():
            # OpenRazer is the only backend
            return self.appdata.menubar.openrazer.restart_daemon()

        if device.supported:
            image = common.get_icon("empty", "openrazer")
            title = device.name
            desc = self._("This supported device wasn't detected by the OpenRazer daemon (v1.0.0)").replace("1.0.0", self.middleman.get_backend("openrazer").version)
            buttons = [
                {
                    "label": "{0} {1}".format(self._("Restart"), backend_name),
                    "icon_folder": "general",
                    "icon_name": "refresh",
                    "action": _restart_backend
                },
                {
                    "label": self._("Troubleshoot"),
                    "icon_folder": "general",
                    "icon_name": "preferences",
                    "action": self._start_troubleshooter
                },
                {
                    "label": self._("Online Help"),
                    "icon_folder": "general",
                    "icon_name": "external",
                    "action": self._open_openrazer_help_unrecognised
                },
            ]
        else:
            image = common.get_icon("empty", "nodevice")
            title = self._("Unsupported Device") + f": {device.vid}:{device.pid}"
            desc = self._("This device needs support adding to OpenRazer before it can be used in Polychromatic.")
            buttons = [
                {
                    "label": self._("Troubleshoot"),
                    "icon_folder": "general",
                    "icon_name": "preferences",
                    "action": self._start_troubleshooter
                },
                {
                    "label": self._("Online Help"),
                    "icon_folder": "general",
                    "icon_name": "external",
                    "action": self._open_openrazer_help_unsupported
                }
            ]

        self.widgets.populate_empty_state(layout, image, title, desc, buttons)

    def open_bad_device(self, msg1, msg2, exception):
        """
        Show a page to inform the user the device could not be opened. Possibly
        due to a temporary glitch or unsupported feature.
        """
        layout = self.Contents.layout()
        layout.setContentsMargins(0, 0, 0, 0)
        shared.clear_layout(layout)

        def _view_details():
            _ = self._
            self.widgets.open_dialog(self.widgets.dialog_generic, _("Error Details"), msg1, msg2, exception)

        self.widgets.populate_empty_state(layout, common.get_icon("empty", "nobackend"), self._("There was a problem opening this device"), "",
            [
                {
                    "label": self._("View Details"),
                    "icon_folder": "emblems",
                    "icon_name": "software",
                    "action": _view_details
                }
            ])
        self.set_cursor_normal()

    def open_apply_to_all(self):
        """
        Show options that change the device state for all connected devices.
        """
        self.set_cursor_busy()
        layout = self.Contents.layout()
        layout.setContentsMargins(15, 15, 15, 15)
        shared.clear_layout(layout)

        btngrp = QButtonGroup()
        bulk_options = bulkapply.BulkApplyOptions(self.middleman)
        mix_match_msg = self._("Not available for all connected devices")

        def _create_button(option):
            """Creates a button similar to an effect button"""
            button = QToolButton()
            button.setText(option.label)
            button.setIconSize(QSize(40, 40))
            button.setToolButtonStyle(Qt.ToolButtonTextUnderIcon)
            button.setIcon(QIcon(option.icon))
            button.setMinimumHeight(70)
            button.setMinimumWidth(105)
            button.option = option
            btngrp.addButton(button)
            if option.label.find("*") > 0:
                button.setToolTip(mix_match_msg)
            return button

        def _create_group_widget(label, widgets):
            """Nests widgets into a group"""
            group = self.widgets.create_group_widget(label)
            container = QWidget()
            container.setLayout(QFlowLayout())
            container.setContentsMargins(QMargins(30, 6, 15, 0))
            for widget in widgets:
                container.layout().addWidget(widget)
            layout.addWidget(group)
            group.layout().addWidget(container)

        def _add_to_page(label, options):
            """Adds this set of bulk options onto the page"""
            if options:
                buttons = []
                for option in options:
                    buttons.append(_create_button(option))
                _create_group_widget(label, buttons)

        _add_to_page(self._("Brightness"), bulk_options.brightness)
        _add_to_page(self._("Effects"), bulk_options.effects)

        if bulk_options.mix_match:
            mixmatch_label = QLabel("* " + mix_match_msg)
            mixmatch_label.setContentsMargins(QMargins(40, 0, 15, 0))
            mixmatch_label.setDisabled(True)
            layout.addWidget(mixmatch_label)

        _add_to_page(self._("Colours"), bulk_options.colours)

        def _bulk_grp_clicked(button):
            try:
                button.option.apply()
            except Exception as e:
                self._catch_command_error(self.current_device, e)

        btngrp.buttonClicked.connect(_bulk_grp_clicked)
        self.btn_grps["bulk"] = btngrp

        layout.addStretch()
        self.set_cursor_normal()

    def show_device_info(self):
        """
        Opens a dialogue box describing the metadata for the device.
        """
        # Alias
        _ = self._

        # Dialog Controls
        dialog = shared.get_ui_widget(self.appdata, "device-info", QDialog)
        tree = dialog.findChild(QTreeWidget, "DeviceTree")
        btn_close = dialog.findChild(QPushButton, "Close")
        btn_test_matrix = dialog.findChild(QPushButton, "TestMatrix")

        # Dialog Button Icons
        if not self.appdata.system_qt_theme:
            btn_test_matrix.setIcon(self.widgets.get_icon_qt("general", "matrix"))
            btn_close.setIcon(self.widgets.get_icon_qt("general", "close"))

        tree.setColumnWidth(0, 250)
        root = tree.invisibleRootItem()

        # Connect signals
        def _close():
            dialog.accept()

        def _populate_tree():
            self.set_cursor_busy()
            root.takeChildren()

            def mkitem(data, value="", icon=None, disabled=False):
                item = QTreeWidgetItem()
                item.setText(0, data)
                if type(value) in [str, int]:
                    item.setText(1, str(value))
                elif type(value) == bool:
                    if value == True:
                        item.setText(1, _("Yes"))
                        item.setIcon(1, QIcon(common.get_icon("general", "success")))
                    else:
                        item.setText(1, _("No"))
                if disabled:
                    item.setDisabled(True)
                if icon:
                    item.setIcon(1 if value != "" else 0, QIcon(icon))
                return item

            backend = device.backend

            hw = mkitem(_("Hardware"))
            hw.addChild(mkitem(_("Name"), device.name))
            hw.addChild(mkitem(_("Backend"), backend.name, os.path.join(self.paths.data_dir, "img", "logo", backend.logo)))
            if device.vid:
                hw.addChild(mkitem("VID:PID", "{0}:{1}".format(device.vid, device.pid)))
            else:
                hw.addChild(mkitem("VID:PID", None))
            hw.addChild(mkitem(_("Form Factor"), device.form_factor["label"], device.form_factor["icon"]))
            hw.addChild(mkitem(_("Serial"), device.serial))
            if device.firmware_version:
                hw.addChild(mkitem(_("Firmware Version"), device.firmware_version))
            if device.keyboard_layout:
                hw.addChild(mkitem(_("Keyboard Layout"), device.keyboard_layout))
            tree.addTopLevelItem(hw)

            # Software Effects
            cfx = mkitem(_("Custom Effects"))
            fx_supported = True if device.matrix else False
            cfx.addChild(mkitem(_("Supported"), fx_supported, disabled=not fx_supported))
            if device.matrix:
                btn_test_matrix.setDisabled(False)
                dimensions = common.get_plural(device.matrix.rows, _("1 row"), _("2 rows").replace("2", str(device.matrix.rows)))
                dimensions += ", " + common.get_plural(device.matrix.cols, _("1 column"), _("2 columns").replace("2", str(device.matrix.cols)))
                cfx.addChild(mkitem(_("Matrix Dimensions"), dimensions, common.get_icon("general", "matrix")))

                if device.monochromatic:
                    cfx.addChild(mkitem(_("Colour Range"), _("256 colours"), common.get_icon("params", "triple")))
                else:
                    cfx.addChild(mkitem(_("Colour Range"), _("16 million colours"), common.get_icon("tray", "ring")))
            tree.addTopLevelItem(cfx)

            # Battery
            if device.battery:
                battery = mkitem(_("Battery"))
                # TODO: Determine battery icon from a new icon class (reusual)
                if device.battery.is_charging:
                    battery.addChild(mkitem(_("Status"), _("Charging (0%)").replace("0", str(device.battery.percentage))))
                else:
                    battery.addChild(mkitem(_("Status"), _("Discharging (0%)").replace("0", str(device.battery.percentage))))
                battery.addChild(mkitem(_("Type"), _("Removable") if device.battery.is_removable else _("Rechargable")))
                tree.addTopLevelItem(battery)

            # DPI
            if device.dpi:
                dpi = mkitem(_("DPI"))
                dpi.addChild(mkitem(_("Current DPI"), f"({device.dpi.x}, {device.dpi.y})"))
                default_stages = []
                for stage in device.dpi.default_stages:
                    dpi_x, dpi_y = stage
                    default_stages.append(dpi_x if dpi_x == dpi_y else f"({dpi_x},{dpi_y})")
                dpi.addChild(mkitem(_("Default Stages"), ", ".join(map(str, default_stages))))

                if device.dpi.user_stages:
                    stages = [value[0] if value[0] == value[1] else f"({value[0]},{value[1]})" for value in device.dpi.user_stages]
                    dpi.addChild(mkitem(_("Custom Stages"), ", ".join(stages)))
                else:
                    dpi.addChild(mkitem(_("Custom Stages"), self._("(Not set)"), disabled=True))
                dpi.addChild(mkitem(_("Supports Sync?"), device.dpi.can_sync, disabled=not device.dpi.can_sync))
                dpi.addChild(mkitem(_("Minimum"), device.dpi.min))
                dpi.addChild(mkitem(_("Maximum"), device.dpi.max))
                tree.addTopLevelItem(dpi)

            def _get_colour_item(option, colours_required):
                item = mkitem(_("Colours"), str(colours_required))
                for index in range(0, colours_required):
                    colour_hex = option.colours[index]
                    item.addChild(mkitem(_("Input 0").replace("0", str(index + 1)), colour_hex, common.generate_colour_bitmap(self.dbg, colour_hex)))
                return item

            # Zones
            zones = mkitem(_("Zones"))
            exclude_expand = []
            for zone in device.zones:
                zone_item = mkitem(zone.label, "", zone.icon)

                for option in zone.options:
                    option_item = mkitem(option.label, "", option.icon)
                    exclude_expand.append(option_item)

                    option_type = "Unknown"
                    if isinstance(option, Backend.EffectOption):
                        option_type = _("Effect")
                    elif isinstance(option, Backend.ToggleOption):
                        option_type = _("Toggle")
                    elif isinstance(option, Backend.SliderOption):
                        option_type = _("Slider")
                    elif isinstance(option, Backend.MultipleChoiceOption):
                        option_type = _("Dropdown")

                    option_item.addChild(mkitem(_("Option ID"), option.uid))
                    option_item.addChild(mkitem(_("Type"), option_type))
                    option_item.addChild(mkitem(_("Active"), option.active))

                    if option.parameters:
                        param_parent = mkitem(_("Parameters"))
                        for param in option.parameters:
                            param_item = mkitem(param.label, "", param.icon)
                            param_item.addChild(mkitem(_("Parameter ID"), param.data))
                            param_item.addChild(mkitem(_("Active"), param.active))
                            if param.colours_required:
                                param_item.addChild(_get_colour_item(option, param.colours_required))
                            param_parent.addChild(param_item)
                        option_item.addChild(param_parent)

                    if isinstance(option, Backend.SliderOption):
                        option_item.addChild(mkitem(_("Current Value"), f"{str(option.value)} {option.suffix}"))
                        option_item.addChild(mkitem(_("Minimum"), option.min))
                        option_item.addChild(mkitem(_("Maximum"), option.max))
                        option_item.addChild(mkitem(_("Interval"), option.step))

                    if option.colours_required:
                        option_item.addChild(_get_colour_item(option, option.colours_required))

                    zone_item.addChild(option_item)
                    zone_item.setExpanded(False)
                zones.addChild(zone_item)

            tree.addTopLevelItem(zones)
            tree.expandAll()
            for item in exclude_expand:
                item.setExpanded(False)

            dialog.setWindowTitle("{0} - {1}".format(_("Device Information"), device.name))
            self.set_cursor_normal()

        def _test_matrix():
            dialog.accept()
            self._test_device_matrix(device)

        device = self.current_device
        btn_close.clicked.connect(_close)
        btn_test_matrix.clicked.connect(_test_matrix)
        dialog.open()
        _populate_tree()

    def _test_device_matrix(self, device):
        """
        Opens a dialogue box to allow the user to test the individual key
        lighting for their device.
        """
        _ = self._

        # Dialog Controls
        self.dialog = shared.get_ui_widget(self.appdata, "inspect-matrix", QDialog)
        label = self.dialog.findChild(QLabel, "Label")
        table = self.dialog.findChild(QTableWidget, "Matrix")
        btn_close = self.dialog.findChild(QPushButton, "Close")
        cur_pos = self.dialog.findChild(QLabel, "CurrentPosition")

        # Dialog Button Icons
        if not self.appdata.system_qt_theme:
            btn_close.setIcon(self.widgets.get_icon_qt("general", "close"))

        def _close_test():
            self.middleman.replay_active_effect(device)
            self.dialog.accept()

        # Populate table
        for x in range(0, device.matrix.cols):
            table.insertColumn(x)
            header = QTableWidgetItem()
            header.setText(str(x))
            table.setHorizontalHeaderItem(x, header)

        for y in range(0, device.matrix.rows):
            table.insertRow(y)
            header = QTableWidgetItem()
            header.setText(str(y))
            table.setVerticalHeaderItem(y, header)

        def _set_pos():
            indexes = table.selectedIndexes()
            device.matrix.clear()
            for index in indexes:
                device.matrix.set(index.column(), index.row(), 255, 255, 255)
            if len(indexes) == 0:
                cur_pos.setText("")
            elif len(indexes) == 1:
                x = indexes[0].column()
                y = indexes[0].row()
                cur_pos.setText(_("Coordinate: X,Y").replace("X", str(x)).replace("Y", str(y)))
            elif len(indexes) > 1:
                cur_pos.setText(_("Coordinate: (Multiple)"))
            device.matrix.draw()

        # Connect signals and set focus
        btn_close.clicked.connect(_close_test)
        table.itemSelectionChanged.connect(_set_pos)
        table.setFocus()
        table.setCurrentCell(0, 0)
        self.dialog.adjustSize()
        self.dialog.resize(QSize(self.dialog.width() + 5, self.dialog.height()))
        self.dialog.setWindowTitle(self._("Inspect Matrix") + " - " + device.name)
        self.dialog.open()


class SpecialControls(shared.TabData):
    """
    Produces Polychromatic-specific controls for features of this software or
    a fancier interface for hardware settings which would be limited when
    implemented at a backend-level.
    """
    def __init__(self, appdata):
        super().__init__(appdata)

    def create_dpi_control(self, device):
        """
        Creates a sophicated control for setting the DPI X/Y values.

        If a device has fixed requirements, Backend.DeviceItem.DPI shouldn't be used
        and instead be implemented as an MultipleChoiceOption or SliderOption.
        """
        dpi_widget = shared.get_ui_widget(self.appdata, "dpi-control", QWidget)

        grid = dpi_widget.findChild(QTableWidget, "Grid")
        slider_x = dpi_widget.findChild(QSlider, "DPI_X_Slider")
        value_x = dpi_widget.findChild(QLabel, "DPI_X_Value")
        slider_y = dpi_widget.findChild(QSlider, "DPI_Y_Slider")
        value_y = dpi_widget.findChild(QLabel, "DPI_Y_Value")
        slider_lock = dpi_widget.findChild(QToolButton, "LockXY")

        # Button group must be attached to class otherwise GC will destroy it
        self.stage_buttons_group = QButtonGroup()
        self.stage_buttons_group.setExclusive(False)

        if not self.appdata.system_qt_theme:
            slider_lock.setIcon(self.widgets.get_icon_qt("general", "lock"))
            # Button padding is too large for this small button
            slider_lock.setStyleSheet("QToolButton { padding: 8px; }")
            value_x.setStyleSheet("QLabel { padding: 4px 0; }")
            value_y.setStyleSheet("QLabel { padding: 4px 0; }")

        # Set up slider parameters and initial values
        for slider in [slider_x, slider_y]:
            slider.setMinimum(device.dpi.min)
            slider.setMaximum(device.dpi.max)
            slider.setSingleStep(100)
            slider.setPageStep(200)

        slider_x.setValue(device.dpi.x)
        slider_y.setValue(device.dpi.y)
        value_x.setText(str(device.dpi.x))
        value_y.setText(str(device.dpi.y))

        if int(device.dpi.x) == int(device.dpi.y):
            slider_lock.setChecked(True)

        # Update grid when moving sliders
        grid.setStyleSheet("QTableView { background-color: #000; gridline-color: #008000; }")
        grid_squares = 100
        grid_min_size = 2
        grid_max_size = 66

        for value in range(0, grid_squares):
            grid.insertRow(0)
            grid.insertColumn(0)

        def _center_grid_view():
            horizontal_scroll_bar = grid.horizontalScrollBar()
            vertical_scroll_bar = grid.verticalScrollBar()

            center_x = (horizontal_scroll_bar.maximum() - horizontal_scroll_bar.minimum()) // 2
            center_y = (vertical_scroll_bar.maximum() - vertical_scroll_bar.minimum()) // 2

            horizontal_scroll_bar.setValue(center_x)
            vertical_scroll_bar.setValue(center_y)

        def _refresh_grid_size():
            dpi_x = slider_x.value()
            dpi_y = slider_y.value()
            base_size = 50_000
            grid_width = min(int(base_size / dpi_x), grid_max_size)
            grid_width = max(grid_min_size, grid_width)
            grid_height = min(int(base_size / dpi_y), grid_max_size)
            grid_height = max(grid_min_size, grid_height)

            for pos in range(0, grid_squares):
                grid.setRowHeight(pos, grid_width)
                grid.setColumnWidth(pos, grid_height)

            _center_grid_view()

        # Update controls/label while sliding
        def _update_stage_controls():
            for button in self.stage_buttons_group.buttons():
                button.setChecked(slider_x.value() == slider_y.value() == button.dpi_value)

        def _slider_x_moved(value):
            value_x.setText(str(value))
            if slider_lock.isChecked():
                slider_y.setValue(value)
            _update_stage_controls()

        def _slider_y_moved(value):
            value_y.setText(str(value))
            if slider_lock.isChecked():
                slider_x.setValue(value)
            _update_stage_controls()

        slider_x.sliderMoved.connect(_slider_x_moved)
        slider_x.valueChanged.connect(_slider_x_moved)
        slider_y.sliderMoved.connect(_slider_y_moved)
        slider_y.valueChanged.connect(_slider_y_moved)

        # Send request once dropped
        def _slider_dropped():
            self.dbg.stdout(f"{device.name}: Setting DPI to {slider_x.value()}, {slider_y.value()}", self.dbg.action, 1)
            try:
                device.dpi.set(slider_x.value(), slider_y.value())
            except Exception as e:
                DevicesTab._catch_command_error(self, device=device, err=e)

            # Sync state with stages buttons
            for button in self.stage_buttons_group.buttons():
                button.setChecked(slider_x.value() == button.dpi_value[0] and slider_y.value() == button.dpi_value[1])

        # Some mice have a zero Y axis
        if device.name in ["Razer Basilisk Essential"]:
            slider_y.setHidden(True)
            value_y.setHidden(True)
            slider_lock.setHidden(True)

        # TODO: In future, only update after 'mouse drop' or scroll release.
        # Currently it's difficult under stock QSlider. Dragging with mouse only isn't great.
        slider_x.valueChanged.connect(_slider_dropped)
        slider_y.valueChanged.connect(_slider_dropped)
        slider_x.valueChanged.connect(_refresh_grid_size)
        slider_y.valueChanged.connect(_refresh_grid_size)

        # Prepare 'stage' buttons to quickly set default or user-defined stages
        stage_widget = QWidget()
        stage_widget.setLayout(QHBoxLayout())
        stage_widget.layout().setContentsMargins(0, 0, 0, 0)
        stages = device.dpi.default_stages

        # Use custom stages if the user saved them for this device
        # TODO: Should reuse this code
        dpi_file = os.path.join(self.paths.dpi, device.name + ".list")
        if os.path.exists(dpi_file):
            values = []
            with open(dpi_file, "r") as f:
                for line in f.readlines():
                    try:
                        dpi_x, dpi_y = line.split(",")
                        values.append([int(dpi_x), int(dpi_y)])
                    except ValueError:
                        # TODO: Output to stderr
                        pass
            if values:
                stages = values

        def _set_dpi_by_button(button):
            dpi_x, dpi_y = button.dpi_value
            slider_lock.setChecked(dpi_x == dpi_y)
            slider_x.setValue(dpi_x)
            slider_y.setValue(dpi_y)
            button.setChecked(True)

        def _edit_dpi_stages():
            self.dpi_editor = DPIStageEditor(self.appdata, device)

        for index, value in enumerate(stages):
            dpi_x, dpi_y = value
            button = QPushButton()
            button.setCheckable(True)
            if device.dpi.x == dpi_x and device.dpi.y == dpi_y:
                button.setChecked(True)
            button.setText(str(dpi_x) if dpi_x == dpi_y else f"{dpi_x},{dpi_y}")
            button.dpi_value = value

            stage_widget.layout().addWidget(button)
            self.stage_buttons_group.addButton(button)

        self.stage_buttons_group.buttonClicked.connect(_set_dpi_by_button)

        edit_btn = QPushButton()
        edit_btn.setText(self._("Edit"))
        edit_btn.setToolTip(self._("Edit DPI Stages"))
        edit_btn.setIcon(self.widgets.get_icon_qt("general", "edit"))
        edit_btn.clicked.connect(_edit_dpi_stages)
        stage_widget.layout().addWidget(edit_btn)
        stage_widget.layout().addStretch()

        _refresh_grid_size()

        return self.widgets.create_row_widget(self._("DPI"), [stage_widget, dpi_widget], vertical=True)

    def create_mouse_accel_control(self):
        """
        Creates a button that'll either:
          - Open mouse acceleration settings for known desktop environments
          - Or inform the user if desktop environment unknown.

        Due to the diverse range of desktop environments, getting mouse
        acceleration values isn't supported right now.
        """
        # TODO: Move to middleman to use for all mice
        def _get_current_desktop_env():
            try:
                desktop_env = os.environ["XDG_CURRENT_DESKTOP"]

                # Some distros add a prefix/suffix ("ubuntu", "X-")
                if desktop_env.find("GNOME") != -1:
                    return "GNOME"
                elif desktop_env.find("Cinnamon") != -1:
                    return "Cinnamon"
                elif desktop_env in ["KDE", "MATE", "Pantheon", "LXQt"]:
                    return desktop_env
                return "Unknown"
            except KeyError:
                return "Unknown"

        def _get_command():
            desktop_env = _get_current_desktop_env()
            desktop_to_cmd = {
                "GNOME": "gnome-control-center mouse",
                "Cinnamon": "cinnamon-settings mouse",
                "KDE": "systemsettings5 mouse",
                "MATE": "mate-mouse-properties",
                "Pantheon": "io.elementary.switchboard settings://input/mouse",
                "LXQt": "lxqt-config-input",
            }
            try:
                return desktop_to_cmd[desktop_env]
            except KeyError:
                return None

        def _open_mouse_settings():
            command = _get_command()
            if command:
                self.dbg.stdout("Opening: " + command, self.dbg.action, 1)
                subprocess.Popen(command.split(" "))
                return

            # Desktop environment unknown, inform the user for manual action.
            self.widgets.open_dialog(self.widgets.dialog_generic,
                                     self._("Mouse Acceleration"),
                                     self._("This feature is provided by your operating system. Polychromatic doesn't recognise this desktop environment " + \
                                            "to automatically open the settings window for you.\n\n" + \
                                            "Look for 'mouse', 'input' or 'hardware' in your System Settings."))

        widget = QWidget()
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        widget.setLayout(layout)
        button = QPushButton()
        button.setText(self._("Open Mouse Settings"))
        button.setIconSize(QSize(24, 24))
        button.setIcon(QIcon.fromTheme("input-mouse"))
        button.clicked.connect(_open_mouse_settings)
        layout.addWidget(button)
        layout.addStretch()

        return self.widgets.create_row_widget(self._("Acceleration"), [widget])

    def create_programmable_keys_control(self, path):
        """
        Returns a widget that informs the user that this software does not
        have a key remapping solution right now.
        """
        def _open_input_remapper():
            subprocess.Popen([path])

        button = QPushButton()
        button.setText(self._("Open Input Remapper"))
        button.clicked.connect(_open_input_remapper)

        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Preferred)

        return self.widgets.create_row_widget("", [button, spacer])

    def create_openrazer_macro_control(self):
        """
        For OpenRazer keyboards, returns a widget that informs the user how to
        use the macro recording feature provided by openrazer-daemon.
        """
        def _open_macro_dialog():
            self.widgets.open_dialog(self.widgets.dialog_generic,
                self._("About Macro Recording"),
                self._("OpenRazer provides a simple on-the-fly macro recording feature.\n\n" + \
                "1. Press FN+[M] to enter macro mode.\n" + \
                "2. Press the macro key to assign to. Only M1-M5 are supported.\n" + \
                "3. Press the keys in sequence to record.\n" + \
                "4. Press FN+[M] to exit macro mode.\n\n" + \
                "Macros remain in memory until the openrazer-daemon process is stopped. The replay speed will be instantaneous, which can be problematic with some software/games."))

        button = QPushButton()
        button.setText(self._("Usage Instructions"))
        button.clicked.connect(_open_macro_dialog)

        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Preferred)

        return self.widgets.create_row_widget(self._("Macros"), [button, spacer])


class DPIStageEditor(shared.TabData):
    """
    A dialog to edit the DPI stages and save them for quick access or to sync to the hardware buttons.
    """
    def __init__(self, appdata, device=Backend.DeviceItem):
        super().__init__(appdata)
        self.dialog = shared.get_ui_widget(self.appdata, "custom-dpi", q_toplevel=QDialog)
        self.device = device

        # TODO: Refactor later into new class/object
        self.dpi_file = os.path.join(appdata.paths.dpi, device.name + ".list")
        custom_enabled = os.path.exists(self.dpi_file)

        # Labels
        self.device_name = self.dialog.findChild(QLabel, "DeviceName")
        self.device_range = self.dialog.findChild(QLabel, "DeviceDPIRange")
        self.device_icon = self.dialog.findChild(QLabel, "DeviceIcon")
        self.sync_possible = self.dialog.findChild(QLabel, "SyncPossible")
        self.sync_impossible = self.dialog.findChild(QLabel, "SyncImpossible")

        # Input
        self.stages_enabled = self.dialog.findChild(QCheckBox, "StagesEnabled")
        self.stages_table = self.dialog.findChild(QTableWidget, "StagesTable")

        # Buttons
        self.btn_sync = self.dialog.findChild(QPushButton, "SyncNow")
        self.btn_stage_add = self.dialog.findChild(QPushButton, "StageAdd")
        self.btn_stage_del = self.dialog.findChild(QPushButton, "StageDel")
        self.dialog_buttons = self.dialog.findChild(QDialogButtonBox, "buttonBox")
        self.btn_save = self.dialog_buttons.button(QDialogButtonBox.Save)

        # Set icons for controls
        if not self.appdata.system_qt_theme:
            self.btn_sync.setIcon(self.widgets.get_icon_qt("general", "refresh"))
            self.btn_stage_add.setIcon(self.widgets.get_icon_qt("general", "new"))
            self.btn_stage_del.setIcon(self.widgets.get_icon_qt("general", "minus"))

        # Set up UI
        self.device_name.setText(device.name)
        self.device_range.setText(self.appdata._("Up to 123 DPI").replace("123", str(device.dpi.max)))
        shared.set_pixmap_for_label(self.device_icon, shared.get_real_device_image(device.real_image), 48)
        self.toggle_custom_setting(custom_enabled)
        self.sync_possible.setVisible(device.dpi.can_sync == True)
        self.sync_impossible.setVisible(device.dpi.can_sync == False)
        self.btn_sync.setEnabled(device.dpi.can_sync == True)

        # Set up events
        self.stages_enabled.clicked.connect(self.toggle_custom_setting)
        self.btn_sync.clicked.connect(self.sync_now)
        self.btn_save.clicked.connect(self.save_changes)
        self.btn_stage_add.clicked.connect(self.stage_add)
        self.btn_stage_del.clicked.connect(self.stage_del)

        if os.path.exists(self.dpi_file):
            self.load_from_file()

        self.dialog.open()

    def toggle_custom_setting(self, enabled):
        """
        Update the UI when enabling/disabling custom DPI stages. When disabling, default values will be restored.
        """
        self.stages_enabled.setChecked(enabled)
        for widget in [self.stages_table, self.btn_stage_add, self.btn_stage_del]:
            widget.setEnabled(enabled)

        if not enabled:
            self._update_dpi_table(self.device.dpi.default_stages)

    def _update_dpi_table(self, values=[]):
        """
        Replace the data in the table with the specified values.
        """
        while self.stages_table.columnCount() > 0:
            self.stages_table.removeColumn(0)

        for item in values:
            value_x = item[0]
            value_y = item[1]
            last_column = self.stages_table.columnCount()
            self.stages_table.insertColumn(last_column)

            cell = self.stages_table.item(0, last_column)
            if not cell:
                cell = QTableWidgetItem()
                self.stages_table.setItem(0, last_column, cell)
            if value_x == value_y:
                cell.setText(str(value_x))
            else:
                cell.setText(f"{value_x},{value_y}")

    def load_from_file(self):
        """
        Load saved DPI data from disk.
        """
        # TODO: Should reuse this code
        values = []
        with open(self.dpi_file, "r") as f:
            for line in f.readlines():
                try:
                    dpi_x, dpi_y = line.split(",")
                    values.append([int(dpi_x), int(dpi_y)])
                except ValueError:
                    # TODO: Output to stderr
                    pass
        self._update_dpi_table(values)

    def stage_add(self):
        """
        Add another stage to the end of the list.
        """
        self.stages_table.insertColumn(self.stages_table.columnCount())
        self.stages_table.setCurrentCell(0, self.stages_table.columnCount() - 1)

    def stage_del(self):
        """
        Remove the last stage from the end of the list.
        """
        column = self.stages_table.currentColumn()
        if column == -1:
            self.stages_table.removeColumn(self.stages_table.columnCount() - 1)
        else:
            self.stages_table.removeColumn(self.stages_table.currentColumn())

        self.stages_table.setCurrentCell(0, self.stages_table.currentColumn())

    def parse_dpi(self):
        """
        Return a list of valid DPI values from user input. Unless there is a validation error, which will raise ValueError.

        The list will be in the list format expected by DeviceItem.DPI.sync():
        [ [1800,1800], [6500, 4800], [...] ]
        """
        values = []
        for col in range(0, self.stages_table.columnCount()):
            cell = self.stages_table.item(0, col)
            if not cell:
                continue
            data = cell.text()
            if len(data) == 0:
                continue
            data = data.split(",")
            try:
                if len(data) == 1:
                    values.append([int(data[0]), int(data[0])])
                elif len(data) == 2:
                    values.append([int(data[0]), int(data[1])])
            except ValueError:
                # TODO: Show dialog pop up -or- mark invalid cells before sync clicked?
                # ... setting 'sync_possible' label would be overwritten later after successful sync.
                pass

        return sorted(values, key=lambda item: int(item[0]))

    def sync_now(self):
        """
        Sync the current DPI values displayed to the hardware.
        """
        dpi_stages = self.parse_dpi()

        if not dpi_stages:
            # Restore to defaults
            self.toggle_custom_setting(False)
            dpi_stages = self.parse_dpi()

        self.device.dpi.sync(dpi_stages)
        self.sync_possible.setText(self._("DPI stages were successfully synchronised to the hardware."))

    def save_changes(self):
        """
        Save the custom values, save to disk then close the dialog.
        """
        values = self.parse_dpi()

        if self.stages_enabled.isChecked():
            with open(self.dpi_file, "w") as f:
                for value in self.parse_dpi():
                    dpi_x, dpi_y = value
                    f.write(f"{dpi_x},{dpi_y}\n")

        if not self.stages_enabled.isChecked() and os.path.exists(self.dpi_file) or not values:
            os.remove(self.dpi_file)

        self.appdata.tab_devices.open_device(self.device)
